###################################################################
# This file is part of tk7, a utility program for the
# ICOM IC-Q7 walkie-talkie.
# 
#    Copyright (C) 2001, 2002, Bob Parnass
#					AJ9S
# 
# tk7 is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
# 
# tk7 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with tk7; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# 02111-1307  USA
###################################################################


proc MakeGui { } \
{
	global Cht
	global Chvector
	global GlobalParam
	global ReadRadioFlag
	global TemplateSavedFlag

	set Cht ""
	set GlobalParam(TemplateFilename) untitled.tk7

	set TemplateSavedFlag no
	set ReadRadioFlag no

	# Set custom font and colors.

	SetAppearance

	set msg [OpenDevice]
	
	if { $msg != ""} \
		{
		tk_dialog .opnerror "tk7 error" \
			$msg error 0 OK
		exit
		}
	
	
	###############################################################
	# Menu bar along the top edge.
	###############################################################
	set fr_menubar [MakeMenuBar .mb]
	set mf [frame .mainframe]
	set fr_line1 [frame $mf.line1]
	set fr_lim [frame $mf.lim]
	
	set fr_scan [MakeScanFrame $fr_line1.scan]
	set fr_display [MakeDisplayFrame $fr_line1.dis]
	set fr_misc [MakeMiscFrame $fr_line1.omsg]
	set fr_title [MakeTitleFrame $fr_line1.title]
	pack $fr_title -side right -fill y
	pack $fr_scan $fr_display $fr_misc -side left -fill y


	###############################################################
	# VFO controls window
	###############################################################
	toplevel .vfo
	set fr_vfo .vfo.ctls
	frame $fr_vfo -relief groove

	set fr_search [MakeSearchFrame $fr_vfo.search]
	set fr_bandstack [MakeBandStackFrame $fr_vfo.bandstack]

	pack $fr_search $fr_bandstack \
		-side left \
		-fill both -expand true

	pack $fr_vfo
	
	# Prevent user from closing the VFO controls window unless
	# he elects to exit the entire program.
	wm protocol .vfo WM_DELETE_WINDOW {ExitApplication}
	wm title .vfo "tk7 VFO Settings"
	


	###############################################################
	# Memory Bank controls window
	###############################################################

	toplevel .mbank
	set mbank .mbank.ctls
	frame $mbank -relief groove
	wm title .mbank "tk7 Memory Bank"

	set fr_bank [MakeMemoryBankFrame $mbank.bank]

	pack $fr_bank -side left -fill both -expand true
	pack $mbank -side left -fill both -expand true
	
	# Prevent user from closing the Bank window unless
	# he elects to exit the entire program.
	wm protocol .mbank WM_DELETE_WINDOW {ExitApplication}


	###############################################################
	# Secondary controls window
	###############################################################
	toplevel .controls
	set ctls .controls.ctls
	frame $ctls -relief groove

	set fr_callch [MakeCallChFrame $ctls.call]
	set fr_com [MakeCommFrame $ctls.com]

	pack $fr_callch $fr_com -side left -fill both -expand true
	
	
	set Chvector ""
	
	pack $fr_menubar -side top -fill x -pady 3 -padx 3
	pack $fr_line1 -side top -fill x -pady 3 -padx 3
	pack $fr_lim -side top -fill x -pady 3 -padx 3
	
	
	pack $ctls -side top -fill both -expand true -padx 3 -pady 3
	pack .mainframe -side top -fill both -expand true
	
	update idletasks
	
	###############################################################
	#  Ask the window manager to catch the delete window
	#  event.
	###############################################################
	wm protocol . WM_DELETE_WINDOW {ExitApplication}
	
	# Prevent user from shrinking or expanding main window.
	wm minsize . [winfo width .] [winfo height .]
	# wm maxsize . [winfo width .] [winfo height .]
	
	wm protocol .controls WM_DELETE_WINDOW {ExitApplication}
	wm title .controls "tk7 Secondary Controls"
	
	
	# Prevent user from overshrinking or expanding controls window.
	wm minsize .controls [winfo width .controls] [winfo height .controls]
	# wm maxsize .controls [winfo width .controls] [winfo height .controls]
	
	
	# Prevent user from shrinking or expanding window.
	wm minsize .vfo [winfo width .vfo] [winfo height .vfo]
	# wm maxsize .vfo [winfo width .vfo] [winfo height .vfo]


	# Force main window to appear on top by hiding, then
	# then showing it.
	wm withdraw .
	wm deiconify .

	return
}


###################################################################
# Alter color and font appearance based on user preferences.
###################################################################
proc SetAppearance { } \
{
	global GlobalParam

	if {$GlobalParam(Font) != "" } \
		{
		# Designate a custom font for most widgets.
		option add *font $GlobalParam(Font)
		}

	if {$GlobalParam(BackGroundColor) != "" } \
		{
		# Designate a custom background color for most widgets.
		option add *background $GlobalParam(BackGroundColor)
		}

	if {$GlobalParam(ForeGroundColor) != "" } \
		{
		# Designate a custom foreground color for most widgets.
		option add *foreground $GlobalParam(ForeGroundColor)
		}

	if {$GlobalParam(TroughColor) != "" } \
		{
		# Designate a custom slider trough color
		# for most scale widgets.
		option add *troughColor $GlobalParam(TroughColor)
		}

	return
}



##########################################################
# Check if the configuration file exists.
# If it exits, return 1.
#
# Otherwise, prompt the user to select the
# serial port.
##########################################################

proc FirstTimeCheck { Rcfile } \
{
	global AboutMsg
	global GlobalParam
	global Libdir
	global tcl_platform

	if { [file readable $Rcfile] == 1 } \
		{
		return 0
		}

	tk_dialog .about "About tk7" \
		$AboutMsg info 0 OK

	# No readable config file found.
	# Treat this as the first time the user has run the program.

	# Create a new window with radio buttions and
	# an entry field so user can designate the proper
	# serial port.

	set msg "Please identify the serial port to which\n"
	set msg [append msg "your radio is connected."]

	toplevel .serialport
	set sp .serialport

	label $sp.intro -text $msg

	frame $sp.rbframe
	set fr $sp.rbframe

	if { $tcl_platform(platform) == "windows" } \
		{
		# For Windows.
		radiobutton $fr.com1 -text COM1: -variable port \
			-value {COM1:}
		radiobutton $fr.com2 -text COM2: -variable port \
			-value {COM2:} 
		radiobutton $fr.com3 -text COM3: -variable port \
			-value {COM3:} 
		radiobutton $fr.com4 -text COM4: -variable port \
			-value {COM4:} 

		pack $fr.com1 $fr.com2 $fr.com3 $fr.com4 \
			-side top -padx 3 -pady 3 -anchor w

		} \
	else \
		{
		# For unix, mac, etc..
		radiobutton $fr.s0 -text /dev/ttyS0 -variable port \
			-value {/dev/ttyS0} 
		radiobutton $fr.s1 -text /dev/ttyS1 -variable port \
			-value {/dev/ttyS1} 
		radiobutton $fr.s2 -text /dev/ttyS2 -variable port \
			-value {/dev/ttyS2} 
		radiobutton $fr.s3 -text /dev/ttyS3 -variable port \
			-value {/dev/ttyS3} 
		radiobutton $fr.s4 -text /dev/ttyS4 -variable port \
			-value {/dev/ttyS4} 
		radiobutton $fr.s5 -text /dev/ttyS5 -variable port \
			-value {/dev/ttyS5} 

		pack \
			$fr.s0 $fr.s1 $fr.s2 \
			$fr.s3 $fr.s4 $fr.s5 \
			-side top -padx 3 -pady 3 -anchor w

		}

	radiobutton $fr.other -text "other (enter below)" \
		-variable port \
		-value other

	entry $fr.ent -width 30 -textvariable otherport

	pack $fr.other $fr.ent \
		-side top -padx 3 -pady 3 -anchor w

	button $sp.ok -text "OK" \
		-command \
			{ \
			global GlobalParam

			if {$port == "other"} \
				{
				set GlobalParam(Device) $otherport
				} \
			else \
				{
				set GlobalParam(Device) $port
				}
			}

	button $sp.exit -text "Exit" \
		-command { exit }

	pack $sp.intro -side top -padx 3 -pady 3
	pack $fr -side top -padx 3 -pady 3
	pack $sp.ok $sp.exit -side left -padx 3 -pady 3 -expand true



	bind $fr.ent <Key-Return> \
		{
		global GlobalParam
		set GlobalParam(Device) $otherport
		}

	wm title $sp "Select serial port"
	wm protocol $sp WM_DELETE_WINDOW {exit}

	set errorflag true

	while { $errorflag == "true" } \
		{
		tkwait variable GlobalParam(Device)

		if { $tcl_platform(platform) != "unix" } \
			{
			set errorflag false
			break
			}

		# The following tests do not work properly
		# in Windows. That is why we won't perform
		# the serial port tests when running Windows version.

		if { ([file readable $GlobalParam(Device)] != 1) \
			|| ([file writable $GlobalParam(Device)] != 1)}\
			{
			# Device must be readable, writable

			bell
			tk_dialog .badport "Serial port problem" \
				"Serial port problem" error 0 OK
			} \
		else \
			{
			set errorflag false
			}
		}

	destroy $sp
	return 1
}

##########################################################
# ExitApplication
#
# This procedure can do any cleanup necessary before
# exiting the program.
#
# Disable computer control of the radio, then quit.
##########################################################
proc ExitApplication { } \
{
	global Lid
	global Lfilename
	global TemplateSavedFlag
	global ReadRadioFlag

	if { ($ReadRadioFlag == "yes") \
		&&  ($TemplateSavedFlag == "no") } \
		{
		set msg "You did not save the template data"
		append msg " in a file."

		set result [tk_dialog .sav "Warning" \
			$msg \
			warning 0 Cancel Exit ]

		if {$result == 0} \
			{
			return
			}
		}

	SaveSetup
	DisableCControl

	if { [info exists Lfilename] } \
		{
		if { $Lfilename != "" } \
			{
			# Close log file.
			close $Lid
			}
		}
	exit
}


##########################################################
# NoExitApplication
#
# This procedure prevents the user from
# killing the window.
##########################################################
proc NoExitApplication { } \
{

	set response [tk_dialog .quitit "Exit?" \
		"Do not close this window." \
		warning 0 OK ]

	return
}

##########################################################
# Contruct the top row of pulldown menus
##########################################################
proc MakeMenuBar { f } \
{
	global AboutMsg
	global Device
	global FileTypes
	global GlobalParam
	global Pgm
	global Version

	# File pull down menu
	frame $f -relief groove -borderwidth 3

	menubutton $f.file -text "File" -menu $f.file.m \
		-underline 0
	menubutton $f.view -text "View" -menu $f.view.m \
		-underline 0
	menubutton $f.data -text "Data" -menu $f.data.m \
		-underline 0
	menubutton $f.radio -text "Radio" -menu $f.radio.m \
		-underline 0
	menubutton $f.help -text "Help" -menu $f.help.m \
		-underline 0
	
	
	menu $f.view.m
	AddView $f.view.m

	menu $f.data.m
	AddData $f.data.m


	menu $f.help.m

	$f.help.m add command -label "Readme" \
		-underline 0 \
		-command { \
			set helpfile [format "%s/README" $Libdir ]
			set win [textdisplay_create "README"]
			textdisplay_file $win $helpfile
			}
	
	$f.help.m add command -label "Tcl info" \
		-underline 0 \
		-command { \
			tk_dialog .about "Tcl info" \
				[HelpTclInfo] info 0 OK
			}

	$f.help.m add command -label "License" \
		-underline 0 \
		-command { \
			set helpfile [format "%s/COPYING" $Libdir ]
			set win [textdisplay_create "Notice"]
			textdisplay_file $win $helpfile
			}
	
	$f.help.m add command -label "About tk7" \
		-underline 0 \
		-command { \
			tk_dialog .about "About tk7" \
				$AboutMsg info 0 OK
			}
	
	menu $f.file.m -tearoff no

	$f.file.m add command -label "Open ..." \
		-underline 0 \
		-command {OpenTemplate .mainframe}
	
	$f.file.m add command -label "Save" \
		-underline 0 \
		-command {SaveTemplate .mainframe 0}

	$f.file.m add command -label "Save As ..." \
		-underline 0 \
		-command {SaveTemplate .mainframe 1}

	$f.file.m add separator

	set msg "Import memory channels from CSV file ..."
	$f.file.m add command -label $msg \
		-underline 0 \
		-command {\
			ImportCSV .
			}


	$f.file.m add command -label "Export memory channels ..." \
		-underline 0 \
		-command {ExportChannels .mainframe}

	$f.file.m add separator

	$f.file.m add command -label "Exit" \
		-underline 1 \
		-command { ExitApplication}
	
	menu $f.radio.m -tearoff no
	AddRadio $f.radio.m


	pack $f.file $f.view $f.data $f.radio -side left -padx 10
	pack $f.help -side right
	
	update
	return $f
}



proc MakeScrollPane {w x y} {
   frame $w -class ScrollPane -width $x -height $y
   canvas $w.c -xscrollcommand [list $w.x set] -yscrollcommand [list $w.y set]
   scrollbar $w.x -orient horizontal -command [list $w.c xview]
   scrollbar $w.y -orient vertical   -command [list $w.c yview]
   set f [frame $w.c.content -borderwidth 0 -highlightthickness 0]
   $w.c create window 0 0 -anchor nw -window $f
   grid $w.c $w.y -sticky nsew
   grid $w.x      -sticky nsew
   grid rowconfigure    $w 0 -weight 1
   grid columnconfigure $w 0 -weight 1
   # This binding makes the scroll-region of the canvas behave correctly as
   # you place more things in the content frame.
   bind $f <Configure> [list Scrollpane_cfg $w %w %h]
   $w.c configure -borderwidth 0 -highlightthickness 0
   return $f
}
proc Scrollpane_cfg {w wide high} {
   set newSR [list 0 0 $wide $high]
	return
   if {![string equals [$w cget -scrollregion] $newSR]} {
      $w configure -scrollregion $newSR
   }
}



##########################################################
# Add widgets to the view menu
##########################################################
proc AddView { m } \
{
	global GlobalParam


	# Change font.

	if {$GlobalParam(Font) == ""} \
		{
		set msg "Change Font"
		} \
	else \
		{
		set msg [format "Change Font (%s)" $GlobalParam(Font)]
		}

	$m add command -label $msg -command \
		{
		set ft [font_select]
		if {$ft != ""} \
			{
			set GlobalParam(Font) $ft

			set msg "The change will take effect next "
			set msg [append msg "time you start tk7."]

			tk_dialog .wcf "Change Appearance" \
				$msg info 0 OK
			}
		}

	$m add command -label "Restore Original Font" -command \
		{
		set GlobalParam(Font) ""
		set msg "The change will take effect next "
		set msg [append msg "time you start tk7."]

		tk_dialog .wcf "Change Appearance" $msg info 0 OK
		}

	$m add separator

	$m add command -label "Change Panel Color" -command \
		{
		set col [tk_chooseColor -initialcolor #d9d9d9]
		if {$col != ""} \
			{
			set GlobalParam(BackGroundColor) $col

			set msg "The change will take effect next "
			set msg [append msg "time you start tk7."]

			tk_dialog .wcf "Change Appearance" \
				$msg info 0 OK
			}
		}

	$m add command -label "Change Lettering Color" -command \
		{
		set col [tk_chooseColor -initialcolor black]
		if {$col != ""} \
			{
			set GlobalParam(ForeGroundColor) $col

			set msg "The change will take effect next "
			set msg [append msg "time you start tk7."]

			tk_dialog .wcf "Change Appearance" \
				$msg info 0 OK
			}
		}

	$m add command -label "Change Slider Trough Color" -command \
		{
		set col [tk_chooseColor -initialcolor #c3c3c3]
		if {$col != ""} \
			{
			set GlobalParam(TroughColor) $col

			set msg "The change will take effect next "
			set msg [append msg "time you start tk7."]

			tk_dialog .wcf "Change Appearance" \
				$msg info 0 OK
			}
		}

	$m add separator


	# Helpful tips balloons


	$m add  checkbutton \
		-label "Balloon Help Windows" \
                -variable GlobalParam(BalloonHelpWindows) \
                -onvalue on \
                -offvalue off 

	return
}


##########################################################
# Add widgets to the Data menu
##########################################################
proc AddData { m } \
{
	global GlobalParam


	set hint ""
	append hint "The Encode Image operation "
	append hint "is designed for use when "
	append hint "testing tk7."
	balloonhelp_for $m $hint

	$m add command -label "Validate data" \
		-command \
			{
			if {[ValidateData] == 0} \
				{
				tk_dialog .info "Valiate data" \
				"The data is ok." info 0 OK
				}
			}


	$m add command -label "Check for duplicate frequencies" \
		-command { CkDuplicate }

	$m add command -label "Encode Image" \
		-command { \
			if {[ValidateData] == 0} \
				{
				MakeWait
				EncodeImage
				KillWait
				}
			}

	$m add separator

	$m add command -label "Clear All Channels ..." \
		-command { ClearAllChannels }

	$m add command -label "Sort Channels ..." \
		-command { MakeSortFrame }

	return
}



##########################################################
# Add choices to the Radio menu
##########################################################
proc AddRadio { m } \
{
	global GlobalParam
	global Libdir


	$m add command -label "Read from radio ..." \
		-command { \
			Radio2Template .mainframe
			update
			}
	
	$m add command -label "Write to radio ..." \
		-command { \
			Image2Radio .mainframe
			update
			}
	
	
	$m add separator
	$m add command -label "Interrogate radio for model info ..." \
		-command { \
			global GlobalParam

			set s [GetModelInfo]
			binary scan $s "H*" x
			set GlobalParam(RadioVersion) $x
			update
			}
	

#	$m add separator
#
#	$m add radiobutton \
#		-label "USA model" \
#		-variable GlobalParam(WhichModel) \
#		-value 00
#
#	$m add radiobutton \
#		-label "EUR model" \
#		-variable GlobalParam(WhichModel) \
#		-value 14
#
#	$m add radiobutton \
#		-label "SUPER model" \
#		-variable GlobalParam(WhichModel) \
#		-value AD


	$m add separator


	$m add  checkbutton \
		-label "Debug" \
                -variable GlobalParam(Debug) \
                -onvalue "1" \
                -offvalue "0"

	return $m
}


##########################################################
#
# Create a progress gauge widget.
#
#
# From "Effective Tcl/Tk Programming,"
# by Mark Harrison and Michael McLennan.
# Page 125.
#
##########################################################
proc gauge_create {win {color ""}} \
{
	frame $win -class Gauge

	# set len [option get $win length Length]
	set len 300

	canvas $win.display -borderwidth 0 -background white \
		-highlightthickness 0 -width $len -height 20
	pack $win.display -expand yes -padx 10
	if {$color == ""} \
		{
		set color [option get $win color Color]
		}


	$win.display create rectangle 0 0 0 20 \
		-outline "" -fill $color -tags bar
	$win.display create text [expr {0.5 * $len}] 10 \
		-anchor c -text "0%" -tags value
	return $win
}

proc gauge_value {win val} \
{
	if {$val < 0 || $val > 100} \
		{
		error "bad value \"$val\": should be 0-100"
		}
	set msg [format "%.0f%%" $val]
	$win.display itemconfigure value -text $msg

	set w [expr {0.01 * $val * [winfo width $win.display]}]
	set h [winfo height $win.display]
	$win.display coords bar 0 0 $w $h

	update
}

proc MakeWaitWindow {f cnflag color} \
{
	global CancelXfer

	set CancelXfer 0

	frame $f
	button $f.cancel -text Cancel -command {\
		global CancelXfer; set CancelXfer 1; puts "Canceled"}

	gauge_create $f.g PaleGreen
	option add *Gauge.borderWidth 2 widgetDefault
	option add *Gauge.relief sunken widgetDefault
	option add *Gauge.length 300 widgetDefault
	option add *Gauge.color gray widgetDefault

	pack $f.g -expand yes -fill both \
		-padx 10 -pady 10

	if {$cnflag} \
		{
		pack $f.cancel -side top -padx 3 -pady 3
		}

	

	pack $f
	return $f.g
}

##########################################################
#
# Copy data from radio to template image (a lengthy string).
#
##########################################################
proc Radio2Template { f }\
{
	global FileTypes
	global GlobalParam
	global Home
	global MemFreq
	global MemMode
	global ReadRadioFlag


	set msg ""
	append msg "Instructions (read all steps):\n"
	append msg "1) Ensure the radio is connected to your computer"
	append msg " and powered on.\n"


	set result [tk_dialog .info "Read from radio" \
		$msg \
		info 0 OK Cancel ]

	if {$result} \
		{
		return
		}

	# Read memory image from radio.
	if {[ReadImage]} \
		{
		set ReadRadioFlag no

		set msg "Error while reading from radio."
		tk_dialog .error $msg $msg error 0 OK
		return
		}
		
	set GlobalParam(Populated) 1
	DecodeImage

	set msg "Transfer Complete.\n"
	append msg "Look at the radio display "
	append msg "to view a status message."

	tk_dialog .belch "Read IC-Q7" \
		$msg info 0 OK

	set ReadRadioFlag yes

	return
}

##########################################################
# Write memory image to a template file.
##########################################################
proc SaveTemplate { f asflag } \
{
	global GlobalParam
	global TemplateSavedFlag
	global ReadRadioFlag
	global Mimage
	global Nmessages

	if {[string length $Mimage] <= 0} \
		{
		set msg "You must first read template data from"
		append msg " the radio before saving it in a"
		append msg " template file."
		append msg " (Use the Radio menu for reading"
		append msg " from the radio.)"

		tk_dialog .error "No template data" \
			$msg error 0 OK
		return
		}

	set mitypes \
		{
		{"IC-Q7 template files"           {.tk7}     }
		}

	set filename $GlobalParam(TemplateFilename)

	if { ($GlobalParam(TemplateFilename) == "") \
		|| ($asflag) } \
		{
		set filename \
			[Mytk_getSaveFile $f \
			$GlobalParam(MemoryFileDir) \
			.tk7 \
			"Save IC-Q7 data to template file" \
			$mitypes]
		}



	if { $filename != "" }\
		{

		if {[ValidateData]} {return}
		MakeWait
		EncodeImage

		# Truncate memory image to the proper length.
		# We want to ignore the several FF records
		# which may have been appended
		# at the end of the image.

		# set n [expr {($Nmessages * 32) - 1}]
		# set Mimage [string range $Mimage 0 $n]

		KillWait

		set GlobalParam(TemplateFilename) $filename
		SetWinTitle

		set GlobalParam(MemoryFileDir) \
			[ Dirname $GlobalParam(TemplateFilename) ]

		set fid [open $GlobalParam(TemplateFilename) "w"]
		fconfigure $fid -translation binary
		puts -nonewline $fid $Mimage
		close $fid
		set TemplateSavedFlag yes
		}

	return
}


##########################################################
# Read memory image from a template file.
##########################################################
proc OpenTemplate { f } \
{
	global BytesPerMessage
	global GlobalParam
	global Mimage
	global Nmessages

	set mitypes \
		{
		{"IC-Q7 template files"           {.tk7}     }
		{"Goran Vlaski IC-Q7 Programming Utility file" \
			{.q7 .Q7}      }
		}

	set GlobalParam(TemplateFilename) [Mytk_getOpenFile \
		$f $GlobalParam(MemoryFileDir) \
		"Open template file" $mitypes]


	if { $GlobalParam(TemplateFilename) != "" }\
		{
		set GlobalParam(MemoryFileDir) \
			[ Dirname $GlobalParam(TemplateFilename) ]


		if [ catch { open $GlobalParam(TemplateFilename) "r"} fid] \
			{
			# error
			tk_dialog .error "tk7" \
				"Cannot open file $file" \
				error 0 OK
			set GlobalParam(TemplateFilename) ""
			return
			} 
		fconfigure $fid -translation binary


		if { [regexp -nocase {\.q7$} \
			$GlobalParam(TemplateFilename)] } \
			{
			# User wants to read a Goran Valaski
			# IC-Q7 Programming Utility .Q7 file.
			set GlobalParam(TemplateFilename) ""
			set code [ReadQ7File $fid]
			} \
		elseif { [regexp -nocase {\.ic2$} \
			$GlobalParam(TemplateFilename)] } \
			{
			# User wants to read a Butel ARC2 .IC2 file.
			set GlobalParam(TemplateFilename) ""
			set code [ReadIC2File $fid]
			} \
		else \
			{
			# User specified a .tr7 file.
			set nbytes [expr {$Nmessages * $BytesPerMessage / 2}]
			set Mimage [read $fid $nbytes]
			set code 0
			}

		close $fid
		SetWinTitle
		if {$code == 0} \
			{
			set GlobalParam(Populated) 1
			DecodeImage
			}
		}

	return
}


##########################################################
# Import data from a .ICF (ICOM Clone Format) file
##########################################################
proc ImportICF { f }\
{
	global GlobalParam
	global Icf2Hex
	global Mimage


	if {[info exists Mimage] == 0} \
		{
		set msg "You must open a template file\n"
		append msg " or read an image from the radio\n"
		append msg " before importing channels.\n"

		tk_dialog .importinfo "tk7" \
			$msg info 0 OK
		return
		}


	set filetypes \
		{
		{"ICOM clone format files"     {.ICF .icf}     }
		}


	set filename [Mytk_getOpenFile $f \
		$GlobalParam(MemoryFileDir) \
		"Import data from ICF file" $filetypes]

	if {$filename == ""} then {return ""}

	set GlobalParam(MemoryFileDir) [ Dirname $filename ]

	if [ catch { open $filename "r"} fid] \
		{
		# error
		tk_dialog .error "tk7" \
			"Cannot open file $file" \
			error 0 OK

		return
		} 

	# Read entire .ICF file at one time.
	set allicf [read $fid]

	set line ""
	set i 0

	set Mimage ""
	# For each line in the .csv file.
	foreach line [split $allicf "\n" ] \
		{
		update

		incr i

		# Skip the first 2 lines in the file.


		if { $i > 2 } then\
			{
			set nchar [string len $line]

			# for each char in the line

			set buf ""
#			puts -nonewline stderr "$i) "
			for {set j 6} {$j < $nchar} {incr j} \
				{
				set c [string range $line $j $j]
				if {[info exists Icf2Hex($c)] == 0} \
					{
					puts stderr "Error in ICF file."
					break
					}
				set newc $Icf2Hex($c)
				append buf $newc
#			puts -nonewline stderr "$newc"
				}
			# Translate to binary
			set buf [string toupper $buf]
			set pbuf [PackString $buf]
			append Mimage $pbuf
			}
#		puts -nonewline stderr "\n"
		}
		
	set GlobalParam(TemplateFilename) ""
	SetWinTitle
	DecodeImage
	close $fid

	return
}


##########################################################
# Import memory channels from a .csv file
##########################################################
proc ImportCSV { f }\
{
	global GlobalParam
	global MemDuplex
	global MemFreq
	global MemMode
	global MemOffset
	global MemSkip
	global MemStep
	global MemToneCodeR
	global MemToneCodeT
	global MemToneFlag
	global Mimage


	if {[info exists Mimage] == 0} \
		{
		set msg "You must open a template file\n"
		append msg " or read an image from the radio\n"
		append msg " before importing channels.\n"

		tk_dialog .importinfo "tk7" \
			$msg info 0 OK
		return
		}


	set filetypes \
		{
		{"IC-Q7 memory channel files"     {.csv .txt}     }
		}


	set filename [Mytk_getOpenFile $f \
		$GlobalParam(MemoryFileDir) \
		"Import channels" $filetypes]

	if {$filename == ""} then {return ""}

	set GlobalParam(MemoryFileDir) [ Dirname $filename ]

	if [ catch { open $filename "r"} fid] \
		{
		# error
		tk_dialog .error "tk7" \
			"Cannot open file $file" \
			error 0 OK

		return
		} 

	# Read entire .csv file at one time.
	set allchannels [read $fid]

	set line ""
	set i 0

	# For each line in the .csv file.
	foreach line [split $allchannels "\n" ] \
		{
		update

		incr i
		if { $i > 1 } then\
			{
			# Delete double quote characters.
			regsub -all "\"" $line "" bline
			set line $bline

			if {$line == ""} then {continue}
			
			set msg [ParseCsvLine $line]
			if {$msg != ""} \
				{
				set response [ErrorInFile \
					$msg $line $filename]
				if {$response == 0} then {continue} \
				else {ExitApplication}
				}
			}
		}
		
	close $fid

	return
}

proc ParseCsvLine {line} \
{
	global Ctcss
	global GlobalParam
	global MemDuplex
	global MemFreq
	global MemMode
	global MemOffset
	global MemSkip
	global MemStep
	global MemToneCodeR
	global MemToneCodeT
	global MemToneFlag
	global Mode
	global Skip
	global Step

	if {$line == ""} {return error}
	set mlist [split $line ","]

	set n [llength $mlist] 
	set m [ expr {9 - $n} ]

	# Add empty fields to the end of the line
	# if there are too few fields.

	for {set i 0} {$i < $m} {incr i} \
		{
		append line ","
		}

	set mlist [split $line ","]


	set ch [lindex $mlist 0]
	set freq [lindex $mlist 1]
	set mode [lindex $mlist 2]
	set step [lindex $mlist 3]
	set offset [lindex $mlist 4]
	set duplex [lindex $mlist 5]
	set toneflag [lindex $mlist 6]
	set ctcsst [lindex $mlist 7]
	set ctcssr [lindex $mlist 8]
	set skip [lindex $mlist 9]




	if { ($i < 0) || ($i > 199) } \
		{
		return "Invalid channel $i."
		}


	
	if { ($freq < $GlobalParam(LowestFreq)) \
		|| ($freq > $GlobalParam(HighestFreq)) } \
		{
		if {$skip != "hide"} \
			{
			return "Invalid frequency $freq."
			}
		}

	set nmode [string toupper $mode]
	if {$nmode == ""} \
		{
		set nmode NFM
		}
	if { [info exists Mode($nmode)] == 0 } \
		{
		if {$skip != "hide"} \
			{
			return "Invalid mode $mode."
			}
		}

	set nstep $step
	if {$nstep == ""} \
		{
		set nstep 5
		}
	if {[info exists Step($nstep)] == 0 } \
		{
		if {$skip != "hide"} \
			{
			return "Invalid step $step."
			}
		}

	if {$offset == ""} \
		{
		set noffset 0.000
		} \
	else \
		{
		set noffset [format "%.3f" $offset]
		}

	if { ($noffset < 0.0) || ($noffset > 159.995) } \
		{
		return "Invalid offset $offset."
		}

	if {($duplex != "") && ($duplex != "+") && ($duplex != "-")} \
		{
		return "Invalid duplex flag $duplex."
		}


	# Transmit CTCSS code.

	set nctcsst $ctcsst

	if {$ctcsst == ""} \
		{
		set nctcsst 0.0
		} \
	elseif { [regexp {\.} $ctcsst] == 0} \
		{
		# CTCSS code is probably an integer
		# so append .0 to it.
		set nctcsst [format "%s.0" $ctcsst]
		}

	if { [info exists Ctcss($nctcsst)] == 0 } \
		{
		return "Invalid CTCSS code $ctcsst."
		}


	# Receive CTCSS code.

	set nctcssr $ctcssr

	if {$ctcssr == ""} \
		{
		set nctcssr 0.0
		} \
	elseif { [regexp {\.} $ctcssr] == 0} \
		{
		# CTCSS code is probably an integer
		# so append .0 to it.
		set nctcssr [format "%s.0" $ctcssr]
		}

	if { [info exists Ctcss($nctcssr)] == 0 } \
		{
		return "Invalid CTCSS code $ctcssr."
		}

	# Must be null, a space, hide, skip, or pskip to be valid.
	if {($skip != "") && ($skip != " ")} \
		{
		if { [info exists Skip($skip)] == 0 } \
			{
			return "Invalid skip value $skip."
			}
		}

	set MemFreq($ch) $freq
	set MemMode($ch) $nmode
	set MemStep($ch) $nstep
	set MemOffset($ch) $noffset
	set MemDuplex($ch) $duplex
	set MemToneFlag($ch) $toneflag
	set MemToneCodeT($ch) $nctcsst
	set MemToneCodeR($ch) $nctcssr
	set MemSkip($ch) $skip

#	puts stderr "ImportCSV: ch $ch, $MemFreq($ch), skip: $skip"


	return ""
}


##########################################################
# Read memory image from an open Goran Valaski .q7 file
#
# Inputs:
#		fid	-file descriptor
##########################################################
proc ReadQ7File { fid }\
{
	global GlobalParam
	global Mimage
	global Nmessages
	global Pgm


	# Read the first part of .q7 file one record at a time.

	set image ""

	# for {set i 0} {$i < 97} {incr i}
	for {set i 0} {$i < $Nmessages} {incr i}  \
		{
		set line [read $fid 46]
		set len [string length $line]
		if {$len != 46} \
			{
			set msg "$Pgm: "
			append msg "Corruption in .q7 file"
			puts stderr $msg

			tk_dialog .error "tk7" \
				"Corrupted .q7 file" \
				error 0 OK

			return -1
			}

		set cc [string index $line 4]

		if { [string compare -nocase -length 1 $cc \xE4] } \
			{
			set msg "$Pgm: "
			append msg "Corruption in .q7 file"
			puts stderr $msg

			tk_dialog .error "tk7" \
				"Corrupted .q7 file" \
				error 0 OK

			return -1
			}
	

		set pline [PackString [string range $line 5 44]]
		set plen [string length $pline]

		set dbuf [string range $pline 0 18] 
		set cksum [string range $pline 19 19] 
		set ccksum [CalcCheckSum $dbuf]
	
		binary scan $cksum "H*" icksum
		scan $icksum "%x" cksum
	
#		puts stderr [format "CHECKSUM file: %s, calculated: %s\n" \
#				$cksum $ccksum]
	
		# if {$cksum != $ccksum} 
		if {0} \
			{
			set msg [format \
				"%s: error, checksum mismatch, radio: %s, calculated: %s\n" \
				$Pgm $cksum $ccksum]
			Tattle $msg
			tk_dialog .error "Checksum error while reading file" \
				$msg error 0 OK
			return -1
			}


		# Strip off memory address and count bytes.
		set buf [string range $dbuf 3 end]
		# set buf [string range $pline 5 44]

		set abuf "$i) "
		append abuf [DumpBinary $buf]
		# puts stderr $abuf

		append image $buf
		}

	set Mimage $image
	return 0
}


##########################################################
# Read memory image from an open Butel ARC2 .IC2 file
#
# Inputs:
#		fid	-file descriptor
##########################################################
proc ReadIC2File { fid }\
{
	global GlobalParam
	global Mimage
	global Nmessages
	global Pgm


	# Read the first part of .r2 file one line at a time.

	set image ""

	for {set i 0} {$i < $Nmessages} {incr i} \
		{
		set len [gets $fid line]
		if {$len < 40} \
			{
			set msg "$Pgm: "
			append msg "Corruption in ARC .IC2 file"
			puts stderr $msg

			tk_dialog .error "tk7" \
				"Corrupted ARC2 .IC2 file" \
				error 0 OK

			return -1
			}

		set line [string range $line 0 39]
		set pline [binary format "H40" $line]
		set plen [string length $pline]

		set dbuf [string range $pline 0 18] 


		set cksum [string range $pline 19 19] 
		set ccksum [CalcCheckSum $dbuf]
	
		binary scan $cksum "H*" icksum
		scan $icksum "%x" cksum
	
#		puts stderr [format "CHECKSUM file: %s, calculated: %s\n" \
#				$cksum $ccksum]
	
		if {$cksum != $ccksum} \
			{
			set msg [format \
				"%s: error, checksum mismatch, radio: %s, calculated: %s\n" \
				$Pgm $cksum $ccksum]
			Tattle $msg
			tk_dialog .error "Checksum error while reading file" \
				$msg error 0 OK
			return -1
			}


		# Strip off memory address and count bytes.
		set buf [string range $dbuf 3 end]
		# set buf [string range $pline 5 44]

#		set abuf "$i) "
#		append abuf [DumpBinary $buf]
#		puts stderr $abuf

		append image $buf
		}

	set Mimage $image
	return 0
}


##########################################################
# Show memory channels in a window.
##########################################################
proc ShowChannels { f }\
{
	global Chvector
	set Chvector ""

	return
}

##########################################################
# Export memory channels to a .csv file
##########################################################
proc ExportChannels { f }\
{
	global FileTypes
	global GlobalParam
	global Home
	global MemDuplex
	global MemFreq
	global MemMode
	global MemOffset
	global MemSkip
	global MemStep
	global MemToneCodeR
	global MemToneCodeT
	global MemToneFlag
	global Mimage
	global Ofilename


	if { ([info exists Mimage] == 0) \
		|| ([string length $Mimage] <= 0) } \
		{
		set msg "You must read data from the radio"
		append msg " before exporting channels."
		append msg " (See the Radio menu.)"
		tk_dialog .info "tk7" $msg info 0 OK
		return
		}


	set Ofilename [Mytk_getSaveFile $f \
		$GlobalParam(MemoryFileDir) \
		.csv \
		"Export memory channels to .csv file" \
		$FileTypes]


	if {$Ofilename != ""} \
		{
		set GlobalParam(MemoryFileDir) [ Dirname $Ofilename ]

		set fid [open $Ofilename "w"]


		tk_dialog .belch "Export" \
			"Export Complete" info 0 OK

		# Write first line as the field names.
		puts -nonewline $fid [format "Ch,MHz,Mode,Step,"]
		puts -nonewline $fid [format "Offset,Duplex,TSQL,"]
		puts $fid [format "TxCTCSS,RxCTCSS,Skip"]

		for {set ch 0} {$ch < 200} {incr ch} \
			{
			# if {$MemFreq($ch) <= .000001}
			if { ($GlobalParam(ShowHidden) == 0) \
				&& ($MemSkip($ch) == "hide")} \
				{
				continue
				}

			set toneflag $MemToneFlag($ch)
			if {$MemToneFlag($ch) == "off"} \
				{
				set toneflag  ""
				}

			set skip $MemSkip($ch)
			if {$skip == " "} \
				{
				set skip ""
				}

			set s [format "%d,%s,%s,%s,%s,%s,%s,%s,%s,%s" \
				$ch $MemFreq($ch) \
				$MemMode($ch) \
				$MemStep($ch) \
				$MemOffset($ch) \
				$MemDuplex($ch) \
				$toneflag \
				$MemToneCodeT($ch) \
				$MemToneCodeR($ch) \
				$skip ]

			puts $fid $s
			}

		close $fid
		}
	return
}

proc FileExistsDialog { file } \
{
	set result [tk_dialog .fed "Warning" \
		"File $file already exists. Overwrite file?" \
		warning 0 Cancel Overwrite ]

	puts "result is $result"
	return $result
}


##########################################################
# Copy memory image to the radio
##########################################################
proc Image2Radio { f }\
{
	global FileTypes
	global Mimage
	global ReadRadioFlag

	if { ([info exists Mimage] == 0) \
		|| ([string length $Mimage] <= 0)} \
		{
		# No image to write.
		set msg "You must first read template data from"
		append msg " the radio or a file before"
		append msg " writing it to the radio."

		tk_dialog .error "Write to radio" $msg error 0 OK
		}

	if {$ReadRadioFlag == "yes"} \
		{
		# We read an image from the radio.
		# Cannot read from and write to the radio
		# during the same session or else the radio
		# complains. (Reason unknown.)
		#
		# Tell user to save the image file, exit
		# the program, restart the program, read
		# the image file, then write to the radio.

		set msg ""
		append msg "You cannot read from the radio "
		append msg "and write to the radio during the same "
		append msg "session.\n\n"
		append msg "Please:\n"
		append msg "1) Save the memory image in a file,\n"
		append msg "using File --> Save As ...\n"
		append msg "2) Exit this program.\n"
		append msg "3) Restart this program.\n"
		append msg "4) Open the image file you saved "
		append msg "previously, using File --> Open ...\n "
		append msg "5) Then, you can write the image "
		append msg "to the radio."

		tk_dialog .belch "Write blocked warning" \
			$msg warning 0 OK
		return
		}

	if {[ValidateData]} {return}
	MakeWait
	EncodeImage
	KillWait

	set msg ""
	append msg "Instructions:\n"
	append msg "1) Ensure the radio is connected to"
	append msg " your computer and radio power is on.\n"

	set result [tk_dialog .info "Write to radio" \
		$msg \
		info 0 OK Cancel ]

	if {$result} \
		{
		# User canceled the write.
		return
		}

	set wcode [WriteImage]
	if {$wcode == 1} \
		{
		set msg "Error while writing to the radio."
		tk_dialog .error "Write error" $msg error 0 OK
		KillWait
		} \
	elseif {$wcode == 2} \
		{
		set msg "Error, cannot read radio version info."
		tk_dialog .error "Write error" $msg error 0 OK
		KillWait
		} \
	else \
		{
		set msg "Transfer Complete.\n"
		append msg "Look at the radio display "
		append msg "to view a status message."
		tk_dialog .belch "Transfer Complete" \
			$msg info 0 OK
		}

	return
}

###################################################################
# Return 1 if frequency is in range 0 - 2000 exclusive.
###################################################################
proc FreqInRange { f units } \
{
	if {$units == "mhz" } \
		{
		if { $f > 0 && $f < 2000.0 } \
			{
			return 1
			}
		} \
	elseif {$units == "khz" } \
		{
		if { $f > 0 && $f < 2000000.0 } \
			{
			return 1
			}
		}
	return 0
}

###################################################################
# Return 1 if string 's' is a valid frequency.
# Return 0 otherwise.
#
# Units should be khz or mhz
###################################################################
proc CheckFreqValid { s units }\
{
	if {$s == ""} then {return 0}

	# Check for non-digit and non decimal point chars.
	set rc [regexp {^[0-9.]*$} $s]
	if {$rc == 0} then {return 0}


	# All digits.
	set rc [regexp {^[0-9]*$} $s]
	if {$rc == 1} \
		{
		return [FreqInRange $s $units]
		}

	if {$s == "."} then {return 0}

	# Check for Two or more decimal points
	set tmp $s
	set tmp [split $s "."]
	set n [llength $tmp]
	if { $n >= 3 } then {return 0}
	
	return [FreqInRange $s $units]
}


###################################################################
# Set default receiver parameters
###################################################################
proc SetUp { } \
{
	global env
	global GlobalParam
	global RootDir
	global tcl_platform


	if { [regexp "Darwin" $tcl_platform(os) ] } \
		{
		# For Mac OS X.
		set RootDir ":"
		} \
	else \
		{
		set RootDir "/"
		}

	set GlobalParam(Debug) 0
	# set GlobalParam(Device) /dev/ttyS1
	set GlobalParam(Ifilename) {}
	set GlobalParam(MemoryFileDir) $RootDir
	set GlobalParam(PreviousFreq) 0.0

	return
}



###################################################################
# 
# Define receiver parameters before we read the
# global parameter configuration file in case they are missing
# from the configuration file.
# This avoids a tcl error if we tried to refer to an
# undefined variable.
#
# These initial definitions will be overridden with
# definitions from the configuration file.
#
###################################################################

proc PresetGlobals { } \
{
	global GlobalParam
	global Mode
	global Rcfile
	global RootDir
	global tcl_platform

	set GlobalParam(BalloonHelpWindows) on

	set GlobalParam(Attenuator) 0
	set GlobalParam(AutoOff) OFF
	set GlobalParam(AutoRepeater) OFF
	set GlobalParam(BackGroundColor) ""
	set GlobalParam(BankSort) -1
	set GlobalParam(BatterySaver) 1
	set GlobalParam(Beep) 1
	set GlobalParam(ChannelOnly) 0
	set GlobalParam(Debug) 0
	set GlobalParam(Dial) 1MHz
	set GlobalParam(DialAccel) 1
	set GlobalParam(Expand) 1
	set GlobalParam(Font) ""
	set GlobalParam(ForeGroundColor) ""
	set GlobalParam(FreqSkip) 1
	set GlobalParam(Lamp) AUTO
	set GlobalParam(LimitSearch) 0
	set GlobalParam(Lock) 0
	set GlobalParam(LockEffect) NORMAL
	set GlobalParam(MemoryFileDir) $RootDir
	set GlobalParam(Mode) $Mode(NFM)
	set GlobalParam(Monitor) PUSH
	set GlobalParam(Pause) 10
	set GlobalParam(Priority) OFF
	set GlobalParam(PowerSave) 1
	set GlobalParam(RadioVersion) ""
	set GlobalParam(Resume) 2
	set GlobalParam(RITenable) 0
	set GlobalParam(ScanMB) BOTH
	set GlobalParam(SetMenuItem) 0
	set GlobalParam(ShowHidden) 0
	set GlobalParam(Squelch) AUTO
	set GlobalParam(SWenable) 0
	set GlobalParam(TroughColor) ""
	set GlobalParam(TuningStep) 5
	set GlobalParam(VFOSearch) ALL
	set GlobalParam(VFOFreq) 162.4000
	set GlobalParam(WhichModel) 00
	set GlobalParam(WXFreq) 162.55
	set GlobalParam(WXMode) AUTO

	return
}


###################################################################
# Set global variables after reading the global
# configuration file so these settings override
# whatever values were in the configuration file.
###################################################################

proc OverrideGlobals { } \
{
	global env
	global GlobalParam
	global RootDir
	global tcl_platform


	set GlobalParam(BypassAllEncoding) 0
	set GlobalParam(FileVersion) "                "
	set GlobalParam(Ifilename) {}
	set GlobalParam(LowestFreq) .005
	set GlobalParam(HighestFreq) 1599.995
	set GlobalParam(NmsgsRead) 0
	set GlobalParam(Populated) 0
	set GlobalParam(SortBank) 0
	set GlobalParam(SortType) freq
	set GlobalParam(TemplateFilename) {}
	set GlobalParam(UserComment) "                "
	set GlobalParam(UserPort) 0

	# Note on MacOS X:
	# The initial directory passed to the file chooser widget.
	# The problem here is that osx's tcl is utterly busted.
	# The _only_ pathname it accepts is ':' - no other ones work.
	# Now this isn't as bad as you might think because
	# the native macos file selector widget persistantly
	# remembers the last place you opened/saved a file
	# for a particular application. So the logic to
	# remember this is simply redundant on macos anyway...
	# Presumably they'll fix this someday and we can take
	# out the hack.
	# - Ben Mesander

	if { [regexp "Darwin" $tcl_platform(os) ] } \
		{
		# kluge for MacOS X.

		set GlobalParam(LogFileDir) $RootDir
		set GlobalParam(MemoryFileDir) $RootDir

		if {$GlobalParam(Ifilename) != ""} \
			{
			set GlobalParam(Ifilename) $RootDir
			}
		}

	return
}

###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeImage { } \
{
	global MemFreq
	global MemDuplex
	global MemMode
	global MemOffset
	global MemStep
	global MemToneCodeT
	global MemToneFlag

	MakeWait
	update idletasks


	for {set bn 0} {$bn < 2} {incr bn} \
		{
		set ch [expr {$bn * 100}]
		for {set i 0} {$i < 100} {incr i} \
			{
			ZapChannel $ch
			incr ch
			}
		}

	DecodeMisc
	DecodeMemories
	DecodeSearchBanks
	DecodeBandStack
	DecodeCallChannels
	DecodeSW

	update idletasks
	KillWait
	update idletasks
	return
}

###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeMisc { } \
{
	global GlobalParam
	global ImageAddr
	global Mimage
	global Priority
	global RAutoOff
	global RAutoRepeater
	global RDial
	global RBatterySaver
	global RLamp
	global RLockEffect
	global RMode
	global RMonitor
	global RFstep
	global RPause
	global RPriority
	global RResume
	global RScanMB
	global RSquelch
	global RStep




#	# Parse file version
#	scan $ImageAddr(FileVersion) "%x" first
#	set last [expr {$first + 15}]
#	set GlobalParam(FileVersion) \
#		[string range $Mimage $first $last]
#
#	# Parse user comment 
#	scan $ImageAddr(UserComment) "%x" first
#	set last [expr {$first + 15}]
#	set GlobalParam(UserComment) \
#		[string range $Mimage $first $last]



	# Parse RIT enable, on or off flag
	scan $ImageAddr(Status) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(RITenable) [GetBit $byte 1]


	# Parse radio version 
	scan $ImageAddr(WhichModel) "%x" first
	set byte [string range $Mimage $first $first]
	binary scan $byte "H2" GlobalParam(WhichModel)

	# Expanded set mode
	scan $ImageAddr(Expand) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(Expand) [Char2Int $byte]

	# Frequency Skip while scanning
	scan $ImageAddr(FreqSkip) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(FreqSkip) [Char2Int $byte]

	# Channel indication mode
	scan $ImageAddr(ChannelOnly) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(ChannelOnly) [Char2Int $byte]

	# Parse dial acceleration flag
	scan $ImageAddr(DialAccel) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(DialAccel) [Char2Int $byte]


	# Parse power save flag
	scan $ImageAddr(PowerSave) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(PowerSave) [Char2Int $byte]


	# Parse beep tone, on or off flag
	scan $ImageAddr(Beep) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(Beep) [Char2Int $byte]


	# Parse Auto repeater function
	scan $ImageAddr(AutoRepeater) "%x" first
	set s [string range $Mimage $first $first]
	set n [Char2Int $s]
	if { [info exists RAutoRepeater($n)] } \
		{
		set GlobalParam(AutoRepeater) $RAutoRepeater($n)
		} \
	else \
		{
		set GlobalParam(AutoRepeater) OFF
		}


	# Parse Squelch level
	scan $ImageAddr(Squelch) "%x" first
	set s [string range $Mimage $first $first]
	set n [Char2Int $s]
	if { [info exists RSquelch($n)] } \
		{
		set GlobalParam(Squelch) $RSquelch($n)
		} \
	else \
		{
		set GlobalParam(Squelch) OFF
		}


	# Parse Priority Watch
	scan $ImageAddr(Priority) "%x" first
	set s [string range $Mimage $first $first]
	set n [Char2Int $s]
	if { [info exists RPriority($n)] } \
		{
		set GlobalParam(Priority) $RPriority($n)
		} \
	else \
		{
		set GlobalParam(Priority) OFF
		}


	# Parse Bank Scan selection
	scan $ImageAddr(ScanMB) "%x" first
	set s [string range $Mimage $first $first]
	set n [Char2Int $s]
	if { [info exists RScanMB($n)] } \
		{
		set GlobalParam(ScanMB) $RScanMB($n)
		} \
	else \
		{
		set GlobalParam(ScanMB) BOTH
		}



	# Parse Lock function effect
	scan $ImageAddr(LockEffect) "%x" first
	set s [string range $Mimage $first $first]
	set n [Char2Int $s]
	if { [info exists RLockEffect($n)] } \
		{
		set GlobalParam(LockEffect) $RLockEffect($n)
		} \
	else \
		{
		set GlobalParam(LockEffect) NORMAL
		}



	# Parse Auto Power Off
	scan $ImageAddr(AutoOff) "%x" first
	set s [string range $Mimage $first $first]
	set n [Char2Int $s]
	if { [info exists RAutoOff($n)] } \
		{
		set GlobalParam(AutoOff) $RAutoOff($n)
		} \
	else \
		{
		set GlobalParam(AutoOff) OFF
		}


	# Parse Dial
	scan $ImageAddr(DialStep) "%x" first
	set s [string range $Mimage $first $first]
	set n [Char2Int $s]
	if { [info exists RDial($n)] } \
		{
		set GlobalParam(Dial) $RDial($n)
		} \
	else \
		{
		set GlobalParam(Dial) 1MHz
		}

	scan $ImageAddr(Lamp) "%x" first
	set s [string range $Mimage $first $first]
	set n [Char2Int $s]
	if { [info exists RLamp($n)] } \
		{
		set GlobalParam(Lamp) $RLamp($n)
		} \
	else \
		{
		set GlobalParam(Lamp) AUTO
		}


	scan $ImageAddr(Pause) "%x" first
	set s [string range $Mimage $first $first]
	set n [Char2Int $s]
	if { [info exists RPause($n)] } \
		{
		set GlobalParam(Pause) $RPause($n)
		} \
	else \
		{
		set GlobalParam(Pause) 2
		}


	scan $ImageAddr(Resume) "%x" first
	set s [string range $Mimage $first $first]
	set n [Char2Int $s]
	if { [info exists RResume($n)] } \
		{
		set GlobalParam(Resume) $RResume($n)
		} \
	else \
		{
		set GlobalParam(Resume) 2
		}


	# Monitor key
	scan $ImageAddr(Monitor) "%x" first
	set s [string range $Mimage $first $first]
	set n [Char2Int $s]
	if { [info exists RMonitor($n)] } \
		{
		set GlobalParam(Monitor) $RMonitor($n)
		} \
	else \
		{
		set GlobalParam(Monitor) PUSH
		}


	return
}


###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
# NOTES:
#	Each memory channel is represented by 8 consecutive bytes.
#	The first 3 bytes contain the frequency digits in hex.
#	The most significant nibble in the first byte is 0-E.
#	Frequencies of 1000 MHz and higher start with a letter.
#
#	The fourth byte is a little strange. It contains both the
#	simplex/duplex flag and part of the offset
#
#	
###################################################################
proc DecodeMemories { } \
{
	global Mimage
	global MemDuplex
	global MemFreq
	global MemMode
	global MemOffset
	global MemSkip
	global MemStep
	global MemToneCodeR
	global MemToneCodeT
	global MemToneFlag
	global ImageAddr
	global RCtcss
	global Mode
	global RMode
	global RSkip
	global RStep
	global ToneFlag


	# Parse memory channel frequencies.
	scan $ImageAddr(MemoryFreqs) "%x" first

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set last [expr {$first + 2}]
		set s [string range $Mimage $first $last]

		set f [BCD2Freq3 $s]

		set MemFreq($ch) $f
		incr first 8
		incr last 8
		}

	# Parse memory channel offset frequencies.
	scan $ImageAddr(MemoryOffset) "%x" first

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set last [expr {$first + 2}]
		set s [string range $Mimage $first $last]

		set f [BCD2Offset $s]

		set MemOffset($ch) $f
		incr first 8
		incr last 8
		}



	# Parse memory CTCSS tone code.
	scan $ImageAddr(MemoryToneCode) "%x" first
	set next $first
	incr next

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set byte0 [string range $Mimage $first $first]
		set byte1 [string range $Mimage $next $next]


		set n [GetBitField $byte0 0 5]
		# fix me
		if { [info exists RCtcss($n)] } \
			{
			set MemToneCodeR($ch) $RCtcss($n)
			} \
		else \
			{
			set MemToneCodeR($ch) "0.0"
			}

		set n0 [GetBitField $byte0 6 7]
		set n1 [GetBitField $byte1 0 3]
		set n [expr {($n0 * 16) + $n1}]

		if { [info exists RCtcss($n)] } \
			{
			set MemToneCodeT($ch) $RCtcss($n)
			} \
		else \
			{
			set MemToneCodeT($ch) "0.0"
			}

		incr first 8
		incr next 8
		}


	# Parse memory channel step size.
	scan $ImageAddr(MemorySteps) "%x" first

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set byte [string range $Mimage $first $first]


		set n [GetBitField $byte 4 7]
		if { [info exists RStep($n)] } \
			{
			set MemStep($ch) $RStep($n)
			} \
		else \
			{
			set MemStep($ch) ?
			}

		incr first 8
		incr last 8
		}


#
#	# Parse flag byte.
#	# If it is xFF, the channel is empty so
#	# force the frequency to be 0 MHz.
#
#	scan $ImageAddr(MemoryToneFlag) "%x" first
#
#	for {set ch 0} {$ch < 200} {incr ch} \
#		{
#		set byte [string range $Mimage $first $first]
#		if { [string compare -nocase -length 1 \
#			$byte \xff] == 0 } \
#			{
#			set MemFreq($ch) 0.0
#			}
#
#		incr first
#		incr last
#		}


	# Parse memory channel tone flag.
	scan $ImageAddr(MemoryToneFlag) "%x" first

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set byte [string range $Mimage $first $first]

		set n [GetBitField $byte 0 1]
		set MemToneFlag($ch) $ToneFlag($n)

		incr first
		incr last
		}


	# Parse memory channel duplex/simplex flag.
	scan $ImageAddr(MemoryDuplex) "%x" first

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set byte [string range $Mimage $first $first]
		set n [GetBitField $byte 2 3]

		if {$n == 0}\
			{
			set MemDuplex($ch) ""
			} \
		elseif {$n == 1} \
			{
			set MemDuplex($ch) ""
			} \
		elseif {$n == 2} \
			{
			set MemDuplex($ch) "-"
			} \
		else \
			{
			set MemDuplex($ch) "+"
			}

		incr first
		incr last
		}



	# Parse skip field.
	scan $ImageAddr(MemorySkip) "%x" first

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set byte [string range $Mimage $first $first]


		# set n [GetBitField $byte 0 1]
		set n [GetBitField $byte 4 5]
		if { [info exists RSkip($n)] } \
			{
			set MemSkip($ch) $RSkip($n)
			} \
		else \
			{
			set MemSkip($ch) " "
			}

		incr first
		incr last
		}




	# Parse memory channel mode.
	scan $ImageAddr(MemoryModes) "%x" first

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set byte [string range $Mimage $first $first]

		set a [GetBit $byte 6]
		set b [GetBit $byte 7]

		set n [expr {$a + $a + $b}]
		set MemMode($ch) $RMode($n)


		incr first
		incr last
		}


	# Force zero freq channels to be hidden, and
	# force hidden frequencies to be zero.
	for {set ch 0} {$ch < 200} {incr ch} \
		{
		if {$MemFreq($ch) < .001} \
			{
# puts stderr "hiding channel $ch $MemFreq($ch)"
			set MemSkip($ch) hide
			}
		if {$MemSkip($ch) == "hide"} \
			{
# puts stderr "hiding channel $ch $MemFreq($ch) due to skip"
			ZapChannel $ch
			}
		} 

	return
}


###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeSearchBanks { } \
{
	global LimitScan
	global Mimage
	global MemMode
	global ImageAddr


	# Parse lower limit channels.
	scan $ImageAddr(SearchChanFirst) "%x" first

	for {set bn 0} {$bn < 20} {incr bn} \
		{
		set last $first
		set s [string range $Mimage $first $last]

		set f [Char2Int $s]

		set LimitScan($bn,lower) $f
		incr first 2
		}



	# Parse upper limit channels.
	scan $ImageAddr(SearchChanSecond) "%x" first

	for {set bn 0} {$bn < 20} {incr bn} \
		{
		set last $first
		set s [string range $Mimage $first $last]

		set f [Char2Int $s]

		set LimitScan($bn,upper) $f
		incr first 2
		}

	return
}


###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeBandStack { } \
{
	global BandStack
	global Mimage
	global ImageAddr
	global RCtcss
	global Mode
	global RMode
	global RSkip
	global RStep
	global ToneFlag


	# Parse band stack frequencies.
	scan $ImageAddr(BandStackFreq) "%x" first

	for {set bn 0} {$bn < 8} {incr bn} \
		{
		set last [expr {$first + 2}]
		set s [string range $Mimage $first $last]

		set f [BCD2Freq3 $s]

		set BandStack($bn,freq) $f
		incr first 8
		incr last 8
		}


	# Parse band stack offset frequencies.
	scan $ImageAddr(BandStackOffset) "%x" first

	for {set bn 0} {$bn < 8} {incr bn} \
		{
		set last [expr {$first + 2}]
		set s [string range $Mimage $first $last]

		set f [BCD2Offset $s]

		set BandStack($bn,offset) $f
		incr first 8
		incr last 8
		}



	# Parse memory CTCSS tone code.
	scan $ImageAddr(BandStackToneCode) "%x" first
	set next $first
	incr next

	for {set bn 0} {$bn < 8} {incr bn} \
		{
		set byte0 [string range $Mimage $first $first]
		set byte1 [string range $Mimage $next $next]


		set n [GetBitField $byte0 0 5]
		# fix me
		if { [info exists RCtcss($n)] } \
			{
			set BandStack($bn,tonecoder) $RCtcss($n)
			} \
		else \
			{
			set BandStack($bn,tonecoder) "0.0"
			}

		set n0 [GetBitField $byte0 6 7]
		set n1 [GetBitField $byte1 0 3]
		set n [expr {($n0 * 16) + $n1}]

		if { [info exists RCtcss($n)] } \
			{
			set BandStack($bn,tonecodet) $RCtcss($n)
			} \
		else \
			{
			set BandStack($bn,tonecodet) "0.0"
			}

		incr first 8
		incr next 8
		}


	# Parse band stack step size.
	scan $ImageAddr(BandStackSteps) "%x" first

	for {set bn 0} {$bn < 8} {incr bn} \
		{
		set byte [string range $Mimage $first $first]


		set n [GetBitField $byte 4 7]
		if { [info exists RStep($n)] } \
			{
			set BandStack($bn,step) $RStep($n)
			} \
		else \
			{
			set BandStack($bn,step) ?
			}

		incr first 8
		}



	# Parse flag byte.
	# If it is xFF, the channel is empty so
	# force the frequency to be 0 MHz.

	scan $ImageAddr(BandStackToneFlag) "%x" first

	for {set bn 0} {$bn < 8} {incr bn} \
		{
		set byte [string range $Mimage $first $first]
		if { [string compare -nocase -length 1 \
			$byte \xff] == 0 } \
			{
			set BandStack($bn,freq) 0.0
			}

		incr first
		}


	# Parse band stack tone flag.
	scan $ImageAddr(BandStackToneFlag) "%x" first

	for {set bn 0} {$bn < 8} {incr bn} \
		{
		set byte [string range $Mimage $first $first]

		set n [GetBitField $byte 0 1]
		set BandStack($bn,toneflag) $ToneFlag($n)

		incr first
		}


	# Parse memory channel duplex/simplex flag.
	scan $ImageAddr(BandStackDuplex) "%x" first

	for {set bn 0} {$bn < 8} {incr bn} \
		{
		set byte [string range $Mimage $first $first]
		set n [GetBitField $byte 2 3]

		if {$n == 0}\
			{
			set BandStack($bn,duplex) ""
			} \
		elseif {$n == 1} \
			{
			set BandStack($bn,duplex) ""
			} \
		elseif {$n == 2} \
			{
			set BandStack($bn,duplex) "-"
			} \
		else \
			{
			set BandStack($bn,duplex) "+"
			}

		incr first
		}



	# Parse band stack skip field.
	scan $ImageAddr(BandStackSkip) "%x" first

	for {set bn 0} {$bn < 8} {incr bn} \
		{
		set byte [string range $Mimage $first $first]


		set n [GetBitField $byte 4 5]
		if { [info exists RSkip($n)] } \
			{
			set BandStack($bn,skip) $RSkip($n)
			} \
		else \
			{
			set BandStack($bn,skip) " "
			}

		incr first
		}


	# Parse band stack mode.
	scan $ImageAddr(BandStackModes) "%x" first

	for {set bn 0} {$bn < 8} {incr bn} \
		{
		set byte [string range $Mimage $first $first]

		set n [GetBitField $byte 6 7]
		set BandStack($bn,mode) $RMode($n)

		incr first
		}

	return
}


###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeCallChannels { } \
{
	global CallChannel
	global Mimage
	global ImageAddr
	global RCtcss
	global RMode
	global RSkip
	global RStep
	global ToneFlag


	# Parse call ch frequencies.
	scan $ImageAddr(CallChannelFreq) "%x" first

	for {set bn 0} {$bn < 2} {incr bn} \
		{
		set last [expr {$first + 2}]
		set s [string range $Mimage $first $last]

		set f [BCD2Freq3 $s]

		set CallChannel($bn,freq) $f
		incr first 8
		incr last 8
		}


	# Parse call channel offset frequencies.
	scan $ImageAddr(CallChannelOffset) "%x" first

	for {set bn 0} {$bn < 2} {incr bn} \
		{
		set last [expr {$first + 2}]
		set s [string range $Mimage $first $last]

		set f [BCD2Offset $s]

		set CallChannel($bn,offset) $f
		incr first 8
		incr last 8
		}



	# Parse CTCSS tone code.
	scan $ImageAddr(CallChannelToneCode) "%x" first
	set next $first
	incr next

	for {set bn 0} {$bn < 2} {incr bn} \
		{
		set byte0 [string range $Mimage $first $first]
		set byte1 [string range $Mimage $next $next]


		set n [GetBitField $byte0 0 5]
		# fix me
		if { [info exists RCtcss($n)] } \
			{
			set CallChannel($bn,tonecoder) $RCtcss($n)
			} \
		else \
			{
			set CallChannel($bn,tonecoder) "0.0"
			}

		set n0 [GetBitField $byte0 6 7]
		set n1 [GetBitField $byte1 0 3]
		set n [expr {($n0 * 16) + $n1}]

		if { [info exists RCtcss($n)] } \
			{
			set CallChannel($bn,tonecodet) $RCtcss($n)
			} \
		else \
			{
			set CallChannel($bn,tonecodet) "0.0"
			}

		incr first 8
		incr next 8
		}


	# Parse step size.
	scan $ImageAddr(CallChannelSteps) "%x" first

	for {set bn 0} {$bn < 2} {incr bn} \
		{
		set byte [string range $Mimage $first $first]


		set n [GetBitField $byte 4 7]
		if { [info exists RStep($n)] } \
			{
			set CallChannel($bn,step) $RStep($n)
			} \
		else \
			{
			set CallChannel($bn,step) ?
			}

		incr first 8
		incr last 8
		}



	# Parse flag byte.
	# If it is xFF, the channel is empty so
	# force the frequency to be 0 MHz.

	scan $ImageAddr(CallChannelToneFlag) "%x" first

	for {set bn 0} {$bn < 2} {incr bn} \
		{
		set byte [string range $Mimage $first $first]
		if { [string compare -nocase -length 1 \
			$byte \xff] == 0 } \
			{
			set CallChannel($bn,freq) 0.0
			}

		incr first
		}


	# Parse tone flag.
	scan $ImageAddr(CallChannelToneFlag) "%x" first

	for {set bn 0} {$bn < 2} {incr bn} \
		{
		set byte [string range $Mimage $first $first]

		set n [GetBitField $byte 0 1]
		set CallChannel($bn,toneflag) $ToneFlag($n)

		incr first
		}


	# Parse duplex/simplex flag.
	scan $ImageAddr(CallChannelDuplex) "%x" first

	for {set bn 0} {$bn < 2} {incr bn} \
		{
		set byte [string range $Mimage $first $first]
		set n [GetBitField $byte 2 3]

		if {$n == 0}\
			{
			set CallChannel($bn,duplex) ""
			} \
		elseif {$n == 1} \
			{
			set CallChannel($bn,duplex) ""
			} \
		elseif {$n == 2} \
			{
			set CallChannel($bn,duplex) "-"
			} \
		else \
			{
			set CallChannel($bn,duplex) "+"
			}

		incr first
		}



	# Parse skip field.
	scan $ImageAddr(CallChannelSkip) "%x" first

	for {set bn 0} {$bn < 2} {incr bn} \
		{
		set byte [string range $Mimage $first $first]


		set n [GetBitField $byte 4 5]
		if { [info exists RSkip($n)] } \
			{
			set CallChannel($bn,skip) $RSkip($n)
			} \
		else \
			{
			set CallChannel($bn,skip) " "
			}

		incr first
		}


	# Parse mode.
	scan $ImageAddr(CallChannelModes) "%x" first

	for {set bn 0} {$bn < 2} {incr bn} \
		{
		set byte [string range $Mimage $first $first]

		set n [GetBitField $byte 6 7]
		set CallChannel($bn,mode) $RMode($n)

		incr first
		}

	return
}


###################################################################
# Decode a 3 byte BCD frequency.
#
# Returns:	frequency in MHz
###################################################################

proc BCD2Freq3 { s } \
{
	global GlobalParam
	

	# Note: Icom packs two digits per byte, one per nibble.
	# An important exception is the most significant nibble
	# in the most significant byte.  That nibble can be
	# 0-9 or a-f.
	# a-f means 10-15.

	# Frequency digit pairs.

	set i 0

	set f1 [string index $s $i]
	binary scan $f1 "H2" f1
	regsub -nocase {a} $f1 10 f1
	regsub -nocase {b} $f1 11 f1
	regsub -nocase {c} $f1 12 f1
	regsub -nocase {d} $f1 13 f1
	regsub -nocase {e} $f1 14 f1
	regsub -nocase {f} $f1 15 f1
	incr i

	set f2 [string index $s $i]
	binary scan $f2 "H2" f2
	incr i

	set f3 [string index $s $i]
	binary scan $f3 "H2" f3
	incr i


	set f [format "%s%s%s" $f1 $f2 $f3]
	set f [string trimleft $f 0]

	if { $f == ""} \
		{
		set f "00000000"
		}

	# Check for non-digit chars.
	set rc [regexp {^[0-9]*$} $f]
	if {$rc == 0} then {set f "00000000"}

	set f [expr {$f/1000.0}]
	set f [ format "%.4f" $f]
	return $f
}


###################################################################
# Decode a 2-1/2 byte BCD frequency offset.
#
# Returns:	frequency in MHz
###################################################################

proc BCD2Offset { s } \
{
	global GlobalParam
	

	# Note: Yaesu packs two digits per byte, one per nibble.
	# An important exception is the least significant nibble
	# in the most significant byte.  That nibble can be
	# 0-9 or a-f.
	# a-f means 10-15.

	# Frequency digit pairs.

	set i 0

	set f1 [string index $s $i]

	# Extract right nibble of most significant byte.
	set f1 [GetBitField $f1 4 7]
	incr i

	set f2 [string index $s $i]
	binary scan $f2 "H2" f2
	incr i

	set f3 [string index $s $i]
	binary scan $f3 "H2" f3
	incr i


	set f [format "%d%s%s" $f1 $f2 $f3]
	set f [string trimleft $f 0]

	if { $f == ""} \
		{
		set f "00000000"
		}

	# Check for non-digit chars.
	set rc [regexp {^[0-9]*$} $f]
	if {$rc == 0} then {set f "00000000"}

	set f [expr {$f/1000.0}]
	set f [ format "%.3f" $f]
	return $f
}



###################################################################
# Encode a frequency offset into Yaesu 2-1/2 byte format.
#
# Usage:	Offset2BCD f d
# Inputs:
#		f	-frequency in MHz (0.005 <= f < 160.0 MHz)
#		dir	-direction; null, -, +
#
# Returns:	3 bytes consisting of 1/2 byte direction and
#			2-1/2 byte BCD Yaesu coded frequency
###################################################################
proc Offset2BCD { f dir } \
{
	global GlobalParam
	
	# Note: Yaesu packs two digits per byte, one per nibble.
	# An important exception is the least significant nibble
	# in the most significant byte.  That nibble can be
	# 0-9 or a-f.
	# a-f means 10-15.

	# Check for non-digit chars.
	set rc [regexp {^[0-9\.]*$} $f]
	if {$rc == 0} {set f 0.0}

	if { ($f >= 160.0) || ($f < 0.005) } \
		{
		set f 0.0
		}

	set z [expr {($f * 1000.0) + .0005}]
	set z [expr {int($z)}]

	set z [format "%06d" $z]
	set s [string range $z 0 1]

	regsub -nocase {10} $s "0a" s
	regsub -nocase {11} $s "0b" s
	regsub -nocase {12} $s "0c" s
	regsub -nocase {13} $s "0d" s
	regsub -nocase {14} $s "0e" s
	regsub -nocase {15} $s "0f" s

	set b1 [binary format "H2" $s]

	# Frequency digit pairs.  Two digits per byte.

	set s [string range $z 2 3]
	set b2 [binary format "H2" $s]

	set s [string range $z 4 5]
	set b3 [binary format "H2" $s]


	if {$dir == "+"} \
		{
		set dn 3
		} \
	elseif {$dir == "-"} \
		{
		set dn 2
		} \
	else \
		{
		set dn 0
		}

	# Set the left nibble according to the
	# simplex/duplex+/duplex- value

	set b1 [SetBitField $b1 0 3 $dn]

	set freqbytes ""
	append freqbytes $b1 $b2 $b3

	return $freqbytes
}


###################################################################
# Encode a frequency into Yaesu 3 byte format.
#
# Input:	frequency in MHz
# Returns:	3 byte BCD Yaesu coded frequency
###################################################################
proc Freq2BCD3 { f } \
{
	global GlobalParam
	
	# Note: Yaesu packs two digits per byte, one per nibble.
	# An important exception is the most significant nibble
	# in the most significant byte.  That nibble can be
	# 0-9 or a-f.
	# a-f means 10-15.

	if {$f < .001} \
		{
		# Frequency is zero MHz so set all bits.
		set s [binary format "H2H2H2" FF FF FF]
		return $s
		}
		

	# Check for non-digit chars.
	set rc [regexp {^[0-9\.]*$} $f]
	if {$rc == 0} {set f 0.0}

	if { ($f > $GlobalParam(HighestFreq)) \
		|| ($f < $GlobalParam(LowestFreq)) } \
		{
		set f 0.0
		}

	set z [expr {($f * 1000.0) + .0005}]
	set z [expr {int($z)}]

	if {$f <= 1000.0} \
		{
		set z [format "%06d" $z]

		set s [string range $z 0 1]
		set b1 [binary format "H2" $s]

		set i 2
		set j 3
		} \
	else \
		{
		set z [format "%07d" $z]
		set s [string range $z 0 1]


		regsub -nocase {10} $s "a" s
		regsub -nocase {11} $s "b" s
		regsub -nocase {12} $s "c" s
		regsub -nocase {13} $s "d" s
		regsub -nocase {14} $s "e" s
		regsub -nocase {15} $s "f" s

		set b1 $s
		append b1 [string range $z 2 2]

		set b1 [binary format "H2" $b1]
		set i 3
		set j 4
		}

	# Frequency digit pairs.  Two digits per byte.

	set s [string range $z $i $j]
	set b2 [binary format "H2" $s]

	incr i 2
	incr j 2
	set s [string range $z $i $j]
	set b3 [binary format "H2" $s]

	set freqbytes ""
	append freqbytes $b1 $b2 $b3

	return $freqbytes
}


###################################################################
# Encode a frequency into Yaesu 3 byte plus format.
#
# Input:	frequency in MHz
# Returns:	4 byte BCD Yaesu coded frequency plus flag/mode byte
#
#	The 4th byte returned will be x00, x20, or x40,
#	depending on whether the freq is a 12.5 or 6.25 kHz
#	splinter or not.
#
###################################################################
proc Freq2BCD3p { f } \
{
	global GlobalParam
	
	# Note: Yaesu packs two digits per byte, one per nibble.
	# An important exception is the most significant nibble
	# in the most significant byte.  That nibble can be
	# 0-9 or a-f.
	# a-f means 10-15.


	# Check for non-digit chars.
	set rc [regexp {^[0-9\.]*$} $f]
	if {$rc == 0} {set f 0.0}

	if { ($f > $GlobalParam(HighestFreq)) \
		|| ($f < $GlobalParam(LowestFreq)) } \
		{
		set f 0.0
		}

	# Compensate for roundoff errors.
	set z [expr {($f * 100000.0) + .000005}]
	set z [expr {int($z)}]

	if {$f < 1000.0} \
		{
		set z [format "%08d" $z]

		set s [string range $z 0 1]
		set b1 [binary format "H2" $s]

		set i 2
		set j 3
		} \
	else \
		{
		set z [format "%09d" $z]
		set s [string range $z 0 1]


		regsub -nocase {10} $s "a" s
		regsub -nocase {11} $s "b" s
		regsub -nocase {12} $s "c" s
		regsub -nocase {13} $s "d" s
		regsub -nocase {14} $s "e" s
		regsub -nocase {15} $s "f" s

		set b1 $s
		append b1 [string range $z 2 2]

		set b1 [binary format "H2" $b1]
		set i 3
		set j 4
		}

	# Frequency digit pairs.  Two digits per byte.

	set s [string range $z $i $j]
	set b2 [binary format "H2" $s]

	incr i 2
	incr j 2
	set s [string range $z $i $j]
	set b3 [binary format "H2" $s]

	set flag50 [regexp {50$} $z]
	set flag25 [regexp {25$} $z]

	if {$flag50} \
		{
		# Frequency is a 12.5 kHz splinter
		set b4 [binary format "H2" 40]
		} \
	elseif {$flag25} \
		{
		# Frequency is a 6.25 kHz splinter
		set b4 [binary format "H2" 20]
		} \
	else \
		{
		set b4 [binary format "H2" 00]
		}

	binary scan $b4 "H2" s

	set freqbytes ""
	append freqbytes $b1 $b2 $b3 $b4


	return $freqbytes
}


###################################################################
# Create widgets for the name of this program.
###################################################################
proc MakeTitleFrame { f }\
{
	global DisplayFontSize 
	global Version

	frame $f -relief flat -borderwidth 3

	# set s [format "tk7 v%s" $Version]
	set s [format "tk7"]

	label $f.lab -text $s \
		-background red \
		-foreground white \
		-relief raised \
		-borderwidth 3 \
		-font $DisplayFontSize 

	set s ""
	append s [format "Version %s\n" $Version]
	append s "Experimental Utility\n"
	append s "for the ICOM IC-Q7A Walkie-Talkie"

	label $f.use -text $s \
		-background black \
		-foreground white \
		-relief raised \
		-borderwidth 3

	pack $f.lab $f.use -side top -padx 0 -pady 0 \
		-fill y -fill x -expand true

	return $f
}


###################################################################
# Create frame for display parameters. 
###################################################################
proc MakeDisplayFrame { f }\
{

	frame $f -relief groove -borderwidth 3
	label $f.lab -text "Display, Keypad Settings" \
		-borderwidth 3
	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3
	MakeDispWidgets  $f.b
	pack $f.b -side top -expand true -fill both

	return $f
}

proc MakeDispWidgets {f} \
{
	global GlobalParam


        label $f.lmonitor -text "Monitor key" -borderwidth 3
	tk_optionMenu $f.monitor GlobalParam(Monitor) PUSH HOLD

        label $f.llamp -text "Lamp" -borderwidth 3
	tk_optionMenu $f.lamp GlobalParam(Lamp) OFF ON AUTO

        label $f.llockeffect -text "Lock function effect" -borderwidth 3
	tk_optionMenu $f.lockeffect GlobalParam(LockEffect) \
		NORMAL NO_SQL NO_VOL ALL


	checkbutton $f.beep -text "Confirmation beep" \
		-variable GlobalParam(Beep) \
		-onvalue 1 -offvalue 0

	checkbutton $f.channelonly -text "Channel indication mode" \
		-variable GlobalParam(ChannelOnly) \
		-onvalue 1 -offvalue 0


	set hint ""
	append hint "When the radio is in Channel Indication mode, "
	append hint "only memory channel numbers are displayed, "
	append hint "and not frequencies.\n"
	append hint "Operation is simplified, but the radio "
	append hint "has limited functionality when in\n"
	append hint "this mode."
	balloonhelp_for $f.channelonly $hint


	grid $f.lmonitor -row 4 -column 0 -sticky w
	grid $f.monitor -row 4 -column 1 -sticky ew
	grid $f.llamp -row 6 -column 0 -sticky w
	grid $f.lamp -row 6 -column 1 -sticky ew
	grid $f.llockeffect -row 8 -column 0 -sticky w
	grid $f.lockeffect -row 8 -column 1 -sticky ew
	grid $f.beep -row 14 -column 0 -sticky w -columnspan 2
	grid $f.channelonly -row 16 -column 0 -sticky w -columnspan 2

	return

}


###################################################################
# Create 25 search banks. 
###################################################################
proc MakeSearchFrame { f }\
{
	global GlobalParam


	frame $f -relief groove -borderwidth 3

	frame $f.rb -relief groove -borderwidth 3
	set r $f.rb


        label $r.ldial -text "Fast dial step" -borderwidth 3
	tk_optionMenu $r.dial GlobalParam(Dial) 100kHz 1MHz 10MHz

	# checkbutton $r.dialaccel -text "Dial acceleration" 
	checkbutton $r.dialaccel -text "" \
		-variable GlobalParam(DialAccel) \
		-onvalue 1 -offvalue 0

        label $r.ldialaccel -text "Dial acceleration" -borderwidth 3

        label $r.lvfosearch -text "VFO Search" -borderwidth 3
	$r.lvfosearch configure -foreground yellow

	tk_optionMenu $r.vfosearch GlobalParam(VFOSearch) \
		BAND ALL \
		PROG0 PROG1 PROG2 PROG3 PROG4 PROG5 \
		PROG6 PROG7 PROG8 PROG9 \
		PROG10 PROG11 PROG12 PROG13 PROG14 PROG15 \
		PROG16 PROG17 PROG18 PROG19 \
		PROG20 PROG21 PROG22 PROG23 PROG24

	grid $r.ldial -row 8 -column 1 -sticky w
	grid $r.dial -row 8 -column 2 -sticky ew
	grid $r.ldialaccel -row 12 -column 1 -sticky w
	grid $r.dialaccel -row 12 -column 2 -sticky e

	# pack $r.lvfosearch $r.vfosearch -side left
	pack $r -side top -padx 3 -pady 3

	label $f.lab -text "\nLimit Search\nChannels" -borderwidth 3
	pack $f.lab -side top -padx 3 -pady 3


	ScrollformCreate $f.b
	pack $f.b -expand yes -fill both

	set w [ScrollFormInterior $f.b]

	for {set i 0} {$i < 20} {incr i} \
		{
		MakeSearchBank $w $i
		}

	set hint ""
	append hint "The channels define "
	append hint "the lower and upper frequency "
	append hint "limits for each limit search bank."
	balloonhelp_for $w $hint
	balloonhelp_for $f.lab $hint

	pack $f.b -side top -fill both -expand true -padx 3 -pady 3

	return $f
}



###################################################################
# Create one a set of widgets for one search bank. 
###################################################################
proc MakeSearchBank { f bn }\
{
	global LimitScan
	global GlobalParam

	set msg [format "%sP" $bn]
	label $f.lab$bn -text $msg -borderwidth 3

	entry $f.lower$bn -width 8 \
		-textvariable LimitScan($bn,lower) \
		-background white 

	entry $f.upper$bn -width 8 \
		-textvariable LimitScan($bn,upper) \
		-background white 

	grid $f.lab$bn -row $bn -column 1
	grid $f.lower$bn -row $bn -column 2
	grid $f.upper$bn -row $bn -column 5

	return $f
}



###################################################################
# Create frame for misc parameters. 
###################################################################
proc MakeMiscFrame { f }\
{

	frame $f -relief groove -borderwidth 3
	label $f.lab -text "Misc. Settings" \
		-borderwidth 3
	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3
	MakeMiscWidgets  $f.b
	pack $f.b -side top -expand true -fill both

	return $f
}


###################################################################
# Create widgets for misc. parameters. 
###################################################################
proc MakeMiscWidgets { f } \
{
	global GlobalParam
	global Priority

	checkbutton $f.expand -text "Expand Set Mode" \
		-variable GlobalParam(Expand) \
		-onvalue 1 -offvalue 0

	set hint ""
	append hint "More Set menus become accessible "
	append hint "in the radio when Expand Set Mode is enabled.\n"
	balloonhelp_for $f.expand $hint

	checkbutton $f.battery -text "Power Save" \
		-variable GlobalParam(PowerSave) -onvalue 1 -offvalue 0

        label $f.lautooff -text "Auto power off (min.)" -borderwidth 3
        tk_optionMenu $f.autooff GlobalParam(AutoOff) \
                OFF 30 60 90 120


        label $f.lautorepeater -text "Auto Repeater function" \
		-borderwidth 3
        tk_optionMenu $f.autorepeater GlobalParam(AutoRepeater) \
                OFF DUPLEX DUPLEX_TONE

        label $f.lsquelch -text "Squelch"  -borderwidth 3
        tk_optionMenu $f.squelch GlobalParam(Squelch) \
               OPEN AUTO 1 2 3 4 5 6 7 8 9

	checkbutton $f.rit -text "RIT" \
		-variable GlobalParam(RITenable) -onvalue 1 -offvalue 0

	checkbutton $f.sw -text "Enable reception below 30 MHz" \
		-variable GlobalParam(SWenable) -onvalue 1 -offvalue 0



	set hint ""
	append hint "The IC-Q7 is not supposed to be able to "
	append hint "tune the shortwave frequencies below 30 MHz.\n"
	append hint "Selecting this option tricks "
	append hint "the radio into being able to tune "
	append hint "below 30 MHz using the VFO.\n\n"
	append hint "NOTES:\n"
	append hint "1) The ability to tune below 30 MHz "
	append hint "becomes disabled once you press the BAND key.\n"
	append hint "2) Cloning software cannot be used to assign "
	append hint "frequencies below 30 MHz to memory channels "
	append hint "because the radio will reject the cloning "
	append hint "data."
	balloonhelp_for $f.sw $hint

        grid $f.expand -row 2 -column 0 -sticky w -columnspan 2
        grid $f.battery -row 4 -column 0 -sticky w -columnspan 2

	grid $f.lautooff  -row 8 -column 0 -sticky w
	grid $f.autooff -row 8 -column 1 -sticky ew

	grid $f.lautorepeater  -row 60 -column 0 -sticky w
	grid $f.autorepeater -row 60 -column 1 -sticky ew

	grid $f.lsquelch  -row 70 -column 0 -sticky w
	grid $f.squelch -row 70 -column 1 -sticky ew

        grid $f.rit -row 80 -column 0 -sticky w -columnspan 2
        grid $f.sw -row 90 -column 0 -sticky w -columnspan 2

	return $f
}



###################################################################
# Create frame for scanning parameters. 
###################################################################
proc MakeScanFrame { f }\
{

	frame $f -relief groove -borderwidth 3
	label $f.lab -text "Scan Settings" \
		-borderwidth 3
	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3
	MakeScanWidgets  $f.b
	pack $f.b -side top -expand true -fill both

	return $f
}


###################################################################
# Create widgets for scanning parameters. 
###################################################################
proc MakeScanWidgets { f } \
{
	global GlobalParam
	global Priority



        label $f.lscanmb -text "Scan Memory Bank" -borderwidth 3
        tk_optionMenu $f.scanmb GlobalParam(ScanMB) \
                BOTH 0 1

	checkbutton $f.freqskip -text "Frequency Skip" \
		-variable GlobalParam(FreqSkip) \
		-onvalue 1 -offvalue 0



	set hint ""
	append hint "When Frequency Skip is enabled, "
	append hint "the radio will skip over pskip frequencies "
	append hint "during VFO scanning."
	balloonhelp_for $f.freqskip $hint


        label $f.lpause -text "Scan Pause (sec.)" -borderwidth 3
        tk_optionMenu $f.pause GlobalParam(Pause) \
                2 4 6 8 10 12 14 16 18 20 HOLD

        label $f.labresume -text "Scan Resume (sec.)" -borderwidth 3
        tk_optionMenu $f.resume GlobalParam(Resume) \
                0 1 2 3 4 5 HOLD

        label $f.lpriority -text "Priority Watch" -borderwidth 3
        tk_optionMenu $f.priority GlobalParam(Priority) \
                OFF ON BELL


	grid $f.lscanmb  -row 12 -column 0 -sticky w
	grid $f.scanmb -row 12 -column 1 -sticky ew

        grid $f.freqskip -row 20 -column 0 -sticky w -columnspan 2

	grid $f.lpause  -row 30 -column 0 -sticky w
	grid $f.pause -row 30 -column 1 -sticky ew

	grid $f.labresume  -row 40 -column 0 -sticky w
	grid $f.resume -row 40 -column 1 -sticky ew

	grid $f.lpriority  -row 50 -column 0 -sticky w
	grid $f.priority -row 50 -column 1 -sticky ew


	return $f
}


###################################################################
# Create frame for Communications parameters. 
###################################################################
proc MakeCommFrame { f }\
{

	frame $f -relief groove -borderwidth 3
	label $f.lab -text "Debugging Information" \
		-borderwidth 3
	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3
	MakeCommWidgets  $f.b

	set hint ""
	append hint "Debugging information fields "
	append hint "are useful for testing tk7. "
	balloonhelp_for $f $hint


	pack $f.b -side top -expand true -fill y

	return $f
}


###################################################################
# Create widgets for Communications params. 
###################################################################
proc MakeCommWidgets { f } \
{
	global GlobalParam

	label $f.labpre -text "Radio Version" -borderwidth 3
	entry $f.pre -width 26 \
		-textvariable GlobalParam(RadioVersion) \
		-background yellow 


	label $f.labnmsgs -text "Number Messages Read" -borderwidth 3
	entry $f.nmsgs -width 5 \
		-textvariable GlobalParam(NmsgsRead) \
		-background yellow 

	checkbutton $f.bypassall -text "Bypass All Encoding" \
		-variable GlobalParam(BypassAllEncoding) \
		-onvalue 1 -offvalue 0



	grid $f.labpre  -row 0 -column 0 -sticky w
	grid $f.pre	-row 0 -column 1 -sticky e

	grid $f.labnmsgs -row 12 -column 0 -sticky w
	grid $f.nmsgs	-row 12 -column 1 -sticky e
	grid $f.bypassall	-row 16 -column 0 -columnspan 2

	return $f
}



###################################################################
# Create widgets for Linked Scan Banks. 
###################################################################
proc MakeLinkedScanFrame { f }\
{
	global GlobalParam

	frame $f -relief sunken -borderwidth 3

#	checkbutton $f.bs -text "Bank\nScan" \
#		-variable GlobalParam(BankScan) \
#		-onvalue 0 -offvalue 1
#
#	$f.bs configure -foreground yellow
#
#	set hint ""
#	set hint [append hint "Bank Scan restricts the radio "]
#	set hint [append hint "to scan a channel bank or "]
#	set hint [append hint "a set of channel banks (choose below) "]
#	set hint [append hint "instead of scanning all banks. "]
#	balloonhelp_for $f.bs $hint
#
#
#	pack $f.bs -side top

	return $f
}


###################################################################
# Create widgets for BandStack Memories. 
###################################################################
proc MakeBandStackFrame { f }\
{

	frame $f -relief groove -borderwidth 3
	label $f.lab -text "Bandstacking Registers" -borderwidth 3
	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3

	for {set i 0} {$i < 8} {incr i} \
		{
		MakeBandStackBank $f.b $i
		}

	pack $f.b -side top -padx 3 -pady 3

	set hint ""
	append hint "Bandstacking registers "
	append hint "remember the last VFO frequency and mode "
	append hint "setting for each band."
	balloonhelp_for $f.b $hint

	return $f
}


###################################################################
# Create one a set of widgets for one BandStack bank. 
###################################################################
proc MakeBandStackBank { f bn }\
{
	global Band
	global BandStack
	global GlobalParam

	set msg [format "%.3f - %.3f" \
		$Band(USA,$bn,low) $Band(USA,$bn,high)]

	label $f.lab$bn -text $msg -borderwidth 3

	entry $f.freq$bn -width 10 \
		-textvariable BandStack($bn,freq) \
		-background white 

	tk_optionMenu $f.bsmodemenu$bn BandStack($bn,mode) \
		NFM WFM AM

	tk_optionMenu $f.bsstep$bn BandStack($bn,step) \
		5 6.25 9 10 12.5 15 20 25 30 50 100


	grid $f.lab$bn -row $bn -column 1 -sticky w
	grid $f.freq$bn -row $bn -column 2
	grid $f.bsmodemenu$bn -row $bn -sticky ew -column 3
	grid $f.bsstep$bn -row $bn -sticky ew -column 4

	return $f
}



###################################################################
# Create widgets for Call Channels. 
###################################################################
proc MakeCallChFrame { f }\
{

	frame $f -relief groove -borderwidth 3
	label $f.lab -text "Call Channels" -borderwidth 3
	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3
	set b $f.b

	label $b.lfreq -text "MHz" -borderwidth 3
	label $b.lmode -text "Mode" -borderwidth 3
	label $b.lstep -text "Step" -borderwidth 3
	label $b.loffset -text "Offset" -borderwidth 3
	label $b.lduplex -text "Duplex" -borderwidth 3
	label $b.ltone -text "Tone" -borderwidth 3
	label $b.lctcsst -text "TxCTCSS" -borderwidth 3
	label $b.lctcssr -text "RxCTCSS" -borderwidth 3

	grid $b.lfreq -row 0 -column 8 -sticky w
	grid $b.lmode -row 0 -column 12 -sticky w
	grid $b.lstep -row 0 -column 16 -sticky w
	grid $b.loffset -row 0 -column 20 -sticky w
	grid $b.lduplex -row 0 -column 24 -sticky w
	grid $b.ltone -row 0 -column 28 -sticky w
	grid $b.lctcsst -row 0 -column 32 -sticky w
	grid $b.lctcssr -row 0 -column 36 -sticky w

	for {set i 0} {$i < 2} {incr i} \
		{
		MakeCallCh $b $i
		}

	pack $b -side top -padx 3 -pady 3


	set hint ""
	append hint "The IC-Q7 can access two Call channels "
	append hint "instantly by pressing the Call key "
	append hint "on the radio keypad."
	balloonhelp_for $f $hint

	return $f
}


###################################################################
# Create one a set of widgets for one Call Channel. 
###################################################################
proc MakeCallCh { f bn }\
{
	global Band
	global CallChannel
	global GlobalParam


	set bnplus1 $bn
	incr bnplus1
	label $f.lab$bn -text $bnplus1 -borderwidth 3

	entry $f.freq$bn -width 10 \
		-textvariable CallChannel($bn,freq) \
		-background white 

	tk_optionMenu $f.modemenu$bn CallChannel($bn,mode) \
		NFM WFM AM

	tk_optionMenu $f.step$bn CallChannel($bn,step) \
		5 6.25 9 10 12.5 15 20 25 30 50 100

	entry $f.offset$bn -width 10 \
		-textvariable CallChannel($bn,offset) \
		-background white 

	tk_optionMenu $f.duplex$bn CallChannel($bn,duplex) \
		"+" "-" " "

	tk_optionMenu $f.toneflag$bn CallChannel($bn,toneflag) \
		" " tone tsql pbeep


	entry $f.tonecodet$bn -width 8 \
		-textvariable CallChannel($bn,tonecodet) \
		-background white 

	entry $f.tonecoder$bn -width 8 \
		-textvariable CallChannel($bn,tonecoder) \
		-background white 

	set row [expr {$bn + 1}]
	grid $f.lab$bn -row $row -column 4 -sticky w
	grid $f.freq$bn -row $row -column 8
	grid $f.modemenu$bn -row $row -sticky ew -column 12
	grid $f.step$bn -row $row -sticky ew -column 16
	grid $f.offset$bn -row $row -sticky ew -column 20
	grid $f.duplex$bn -row $row -sticky ew -column 24
	grid $f.toneflag$bn -row $row -sticky ew -column 28
	grid $f.tonecodet$bn -row $row -sticky ew -column 32
	grid $f.tonecoder$bn -row $row -sticky ew -column 36

	return $f
}



###################################################################
# Encode the information from the data structures into
# the memory image string which can be written to the VR-120.
#
# We don't understand the meaning of all the bytes in
# the memory image.  Therefore, the
# image string must already exist and we will only
# change the bytes which we understand.
#
###################################################################
proc EncodeImage { } \
{
	global GlobalParam
	global Mimage

	if {$GlobalParam(BypassAllEncoding)} \
		{
		# puts stderr "EncodeImage: skip encoding"
		return 0
		}

	# puts stderr "EncodeImage: encoding"
	if { ([info exists Mimage] == 0) } \
		{
		# puts stderr "EncodeImage: image does not exist"
		return error
		}

	set image $Mimage

	set image [EncodeMisc $image]
	if { [string length $image] == 0} {return error}

	set image [EncodeMemories $image]
	if { [string length $image] == 0} {return error}

	set image [EncodeSearchBanks $image]
	if { [string length $image] == 0} {return error}

	set image [EncodeBandStack $image]
	if { [string length $image] == 0} {return error}

	set image [EncodeCallChannels $image]
	if { [string length $image] == 0} {return error}

	set image [EncodeSW $image]
	if { [string length $image] == 0} {return error}

	set Mimage $image

	return 0
}



###################################################################
# Encode misc
# information into a memory image.
###################################################################
proc EncodeMisc { image } \
{
	global AutoOff
	global AutoRepeater
	global BatterySaver
	global Dial
	global Fstep
	global ImageAddr
	global GlobalParam
	global Lamp
	global LockEffect
	global Mode
	global Monitor
	global Pause
	global Priority
	global Resume
	global ScanMB
	global Squelch
	global Step


	# Set RIT Enable bit
	scan $ImageAddr(Status) "%x" first
	set b [string index $image $first]
	set newbyte [AssignBit $b 1 $GlobalParam(RITenable)]
	set image [string replace $image $first $first $newbyte]



	# encode Frequency Skip
	scan $ImageAddr(FreqSkip) "%x" first
	set s [format "%02x" $GlobalParam(FreqSkip)]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]

	# encode Dial Acceleration
	scan $ImageAddr(DialAccel) "%x" first
	set s [format "%02x" $GlobalParam(DialAccel)]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]

	# encode Expanded Set Mode
	scan $ImageAddr(Expand) "%x" first
	set s [format "%02x" $GlobalParam(Expand)]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]


	# encode Channel indication mode flag 
	scan $ImageAddr(ChannelOnly) "%x" first
	set s [format "%02x" $GlobalParam(ChannelOnly)]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]


	# encode Power Saver
	scan $ImageAddr(PowerSave) "%x" first
	set s [format "%02x" $GlobalParam(PowerSave)]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]

	# Pause
	scan $ImageAddr(Pause) "%x" first
	set b [string range $image $first $first]
	set s [format "%02x" $Pause($GlobalParam(Pause))]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]


	# Squelch
	scan $ImageAddr(Squelch) "%x" first
	set b [string range $image $first $first]
	set s [format "%02x" $Squelch($GlobalParam(Squelch))]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]

	# Priority Watch
	scan $ImageAddr(Priority) "%x" first
	set b [string range $image $first $first]
	set s [format "%02x" $Priority($GlobalParam(Priority))]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]


	# Resume
	scan $ImageAddr(Resume) "%x" first
	set b [string range $image $first $first]
	set s [format "%02x" $Resume($GlobalParam(Resume))]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]


	# Monitor flag
	scan $ImageAddr(Monitor) "%x" first
	set b [string range $image $first $first]
	set s [format "%02x" $Monitor($GlobalParam(Monitor))]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]


	# Beep tone flag
	scan $ImageAddr(Beep) "%x" first
	set b [string range $image $first $first]
	set s [format "%02x" $GlobalParam(Beep)]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]



	# encode Auto Repeater function
	scan $ImageAddr(AutoRepeater) "%x" first
	set byte [string index $image $first]
	set n [format "%02x" $AutoRepeater($GlobalParam(AutoRepeater))]
	set newbyte [binary format "H2" $n]
	set image [string replace $image $first $first $newbyte]

	# encode Lock function effect
	scan $ImageAddr(LockEffect) "%x" first
	set byte [string index $image $first]
	set n [format "%02x" $LockEffect($GlobalParam(LockEffect))]
	set newbyte [binary format "H2" $n]
	set image [string replace $image $first $first $newbyte]

	# encode Memory Bank scan selection
	scan $ImageAddr(ScanMB) "%x" first
	set byte [string index $image $first]
	set n [format "%02x" $ScanMB($GlobalParam(ScanMB))]
	set newbyte [binary format "H2" $n]
	set image [string replace $image $first $first $newbyte]


	# encode Auto Power Off
	scan $ImageAddr(AutoOff) "%x" first
	set byte [string index $image $first]
	set n [format "%02x" $AutoOff($GlobalParam(AutoOff))]
	set newbyte [binary format "H2" $n]
	set image [string replace $image $first $first $newbyte]

	# encode Dial fast step
	scan $ImageAddr(DialStep) "%x" first
	set byte [string index $image $first]
	set n [format "%02x" $Dial($GlobalParam(Dial))]
	set newbyte [binary format "H2" $n]
	set image [string replace $image $first $first $newbyte]

	# encode Lamp
	scan $ImageAddr(Lamp) "%x" first
	set byte [string index $image $first]
	set n [format "%02x" $Lamp($GlobalParam(Lamp))]
	set newbyte [binary format "H2" $n]
	set image [string replace $image $first $first $newbyte]


#	# Set VFO/Limit Search bit
#	scan $ImageAddr(FlagByte0) "%x" first
#	set byte [string index $image $first]
#	set newbyte [AssignBit $byte 3 $GlobalParam(LimitSearch)]
#	set image [string replace $image $first $first $newbyte]
#
#	# Fast Tuning Step is stored in one byte
#	scan $ImageAddr(FastTuningStep) "%x" first
#	set s $GlobalParam(FastTuningStep)
#	set s [format "%02x" $Fstep($s)]
#	set b [binary format "H2" $s]
#	set image [string replace $image $first $first $b]


	return $image
}


###################################################################
# Encode Short Wave enable
# information into a memory image.
###################################################################
proc EncodeSW { image } \
{
	global ImageAddr
	global GlobalParam


	scan $ImageAddr(SW) "%x" first

	if {$GlobalParam(SWenable) == 0}  \
		{
		set b [binary format "H2" 03]
		set image [string replace $image $first $first $b]
		return $image
		}
		
	set b [binary format "H2" 15]
	set image [string replace $image $first $first $b]

	scan $ImageAddr(Ch129) "%x" first
	set last [expr {$first + 7}]
	set b [binary format "H2H2H2H2H2H2H2H2" 03 00 00 00 00 00 20 82]
	set image [string replace $image $first $last $b]
	
	scan $ImageAddr(SWvfo) "%x" first
	set last [expr {$first + 7}]
	set b [binary format "H2H2H2H2H2H2H2H2" 49 63 6F 6D 43 6C 6F 6E]
	set image [string replace $image $first $last $b]

	return $image
}



###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeSW { } \
{
	global GlobalParam
	global ImageAddr
	global Mimage


	scan $ImageAddr(SW) "%x" first
	set byte [string range $Mimage $first $first]
	if { [string compare -nocase -length 1 $byte \x15] == 0 } \
		{
		set GlobalParam(SWenable) 1
		} \
	else \
		{
		set GlobalParam(SWenable) 0
		}

	return
}


###################################################################
# Encode the memory channel frequency, mode, and other
# flag information into a memory image.
###################################################################

proc EncodeMemories { image } \
{
	global Ctcss
	global ImageAddr
	global MemDuplex
	global MemFreq
	global MemMode
	global MemOffset
	global MemSkip
	global MemStep
	global MemToneCodeR
	global MemToneCodeT
	global MemToneFlag
	global Mode
	global RToneFlag
	global Skip
	global Step

	global ImageAddr

	# Encode channel frequency.
	scan $ImageAddr(MemoryFreqs) "%x" first
	set last [expr {$first + 2}]

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set b [ Freq2BCD3 $MemFreq($ch) ]

		set image [string replace $image $first $last $b]

		incr first 8
		incr last 8
		}

	# Encode channel offset frequencies.
	# and duplex/simple flags.
	scan $ImageAddr(MemoryOffset) "%x" first
	set last [expr {$first + 2}]

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set b [ Offset2BCD $MemOffset($ch) $MemDuplex($ch) ]

		set image [string replace $image $first $last $b]

		incr first 8
		incr last 8
		}



	# Encode channel rcv CTCSS code.
	scan $ImageAddr(MemoryToneCode) "%x" first
	set next $first
	incr next

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set b0 [ string index $image $first ]
		set b1 [ string index $image $next ]

		# CTCSS rcv code
		set ccode $MemToneCodeR($ch)


		if {$ccode != ""} \
			{
			# If pl code is a whole number
			if { ([string first "." $ccode] < 0) } \
				{
				append ccode .0
				}
			}


		if { [info exists Ctcss($ccode)] == 0 } \
			{
			set nr 0
			} \
		else \
			{
			set nr $Ctcss($ccode)
			}

		set b0 [SetBitField $b0 0 5 $nr]


		# CTCSS xmit code
		set ccode $MemToneCodeT($ch)


		if {$ccode != ""} \
			{
			# If pl code is a whole number
			if { ([string first "." $ccode] < 0) } \
				{
				append ccode .0
				}
			}


		if { [info exists Ctcss($ccode)] == 0 } \
			{
			set nt 0
			} \
		else \
			{
			set nt $Ctcss($ccode)
			}

		set n0 [expr {$nt / 16}]
		set n0 [expr {int($n0)}]
		set b0 [SetBitField $b0 6 7 $n0]

		set n1 [expr {fmod($nt, 16)}]
		set n1 [expr {int($n1)}]
		set b1 [SetBitField $b1 0 3 $n1]

#		puts stderr "$ch T: $MemToneCodeT($ch), r: $MemToneCodeR($ch)"

		set image [string replace $image $first $first $b0]
		set image [string replace $image $next $next $b1]

		incr first 8
		incr next 8
		}

	scan $ImageAddr(MemorySteps) "%x" first

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set b [ string index $image $first ]

		# Tuning Step

		if { [info exists Step($MemStep($ch))] == 0 } \
			{
			set m 0
			} \
		else \
			{
			set m $Step($MemStep($ch))
			}
		set b [SetBitField $b 4 7 $m]
#		puts stderr "EncodeMemories: $ch,  $MemFreq($ch), step: $MemStep($ch), m: $m"

		set image [string replace $image $first $first $b]
		incr first 8
		}


	# Encode channel Skip, Tone Squelch Flag, Duplex, Mode
	scan $ImageAddr(MemorySkip) "%x" first

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set b [ string index $image $first ]

		# Skip type
		if { [info exists Skip($MemSkip($ch))] == 0 } \
			{
			set m 0
			} \
		else \
			{
			set m $Skip($MemSkip($ch))
			}

		set b [SetBitField $b 4 5 $m]

		# Tone Squelch flag
		if { [info exists RToneFlag($MemToneFlag($ch))] } \
			{
			set m $RToneFlag($MemToneFlag($ch))
			} \
		else \
			{
			set m $RToneFlag(off)
			}

		set b [SetBitField $b 0 1 $m]

		# Duplex
		if {$MemDuplex($ch) == "+"} \
			{
			set m 3
			} \
		elseif {$MemDuplex($ch) == "-"} \
			{
			set m 2
			} \
		else \
			{
			set m 0
			}

		set b [SetBitField $b 2 3 $m]


		# Encode Mode
		if { [info exists Mode($MemMode($ch))] } \
			{
			set m $Mode($MemMode($ch))
			} \
		else \
			{
			set m $Mode(NFM)
			}
		set b [SetBitField $b 6 7 $m]

		set image [string replace $image $first $first $b]

		incr first 1
		}


	return $image
}


###################################################################
# Encode the Band Stack Memory channel frequency and mode
# information into a memory image.
###################################################################

proc EncodeBandStack { image } \
{
	global BandStack
	global Ctcss
	global ImageAddr
	global Mode
	global RToneFlag
	global Skip
	global Step

	global ImageAddr

	# Encode frequency.
	scan $ImageAddr(BandStackFreq) "%x" first
	set last [expr {$first + 2}]

	for {set ch 0} {$ch < 8} {incr ch} \
		{
		set b [ Freq2BCD3 $BandStack($ch,freq) ]

		set image [string replace $image $first $last $b]

		incr first 8
		incr last 8
		}

	# Encode offset frequencies.
	# and duplex/simple flags.
	scan $ImageAddr(BandStackOffset) "%x" first
	set last [expr {$first + 2}]

	for {set ch 0} {$ch < 8} {incr ch} \
		{
		set b [ Offset2BCD $BandStack($ch,offset) \
			$BandStack($ch,duplex) ]

		set image [string replace $image $first $last $b]

		incr first 8
		incr last 8
		}



	# Encode channel rcv CTCSS code.
	scan $ImageAddr(BandStackToneCode) "%x" first
	set next $first
	incr next

	for {set ch 0} {$ch < 8} {incr ch} \
		{
		set b0 [ string index $image $first ]
		set b1 [ string index $image $next ]

		# CTCSS rcv code
		set ccode $BandStack($ch,tonecoder)
		if { [info exists Ctcss($ccode)] == 0 } \
			{
			set nr 0
			} \
		else \
			{
			set nr $Ctcss($ccode)
			}

		set b0 [SetBitField $b0 0 5 $nr]


		# CTCSS xmit code
		set ccode $BandStack($ch,tonecodet)
		if { [info exists Ctcss($ccode)] == 0 } \
			{
			set nt 0
			} \
		else \
			{
			set nt $Ctcss($ccode)
			}

		set n0 [expr {$nt / 16}]
		set n0 [expr {int($n0)}]
		set b0 [SetBitField $b0 6 7 $n0]

		set n1 [expr {fmod($nt, 16)}]
		set n1 [expr {int($n1)}]
		set b1 [SetBitField $b1 0 3 $n1]


		set image [string replace $image $first $first $b0]
		set image [string replace $image $next $next $b1]

		incr first 8
		incr next 8
		}

	scan $ImageAddr(BandStackSteps) "%x" first

	for {set ch 0} {$ch < 8} {incr ch} \
		{
		set b [ string index $image $first ]

		# Tuning Step

		if { [info exists Step($BandStack($ch,step))] == 0 } \
			{
			set m 0
			} \
		else \
			{
			set m $Step($BandStack($ch,step))
			}
		set b [SetBitField $b 4 7 $m]

		set image [string replace $image $first $first $b]
		incr first 8
		}


	# Encode Skip, Tone Squelch Flag, Duplex, Mode
	scan $ImageAddr(BandStackSkip) "%x" first

	for {set ch 0} {$ch < 8} {incr ch} \
		{
		set b [ string index $image $first ]

		# Skip type
		if { [info exists Skip($BandStack($ch,skip))] == 0 } \
			{
			set m 0
			} \
		else \
			{
			set m $Skip($BandStack($ch,skip))
			}

		set b [SetBitField $b 4 5 $m]

		# Tone Squelch flag
		if { [info exists RToneFlag($BandStack($ch,toneflag))] } \
			{
			set m $RToneFlag($BandStack($ch,toneflag))
			} \
		else \
			{
			set m $RToneFlag(off)
			}

		set b [SetBitField $b 0 1 $m]

		# Duplex
		if {$BandStack($ch,duplex) == "+"} \
			{
			set m 3
			} \
		elseif {$BandStack($ch,duplex) == "-"} \
			{
			set m 2
			} \
		else \
			{
			set m 0
			}

		set b [SetBitField $b 2 3 $m]


		# Encode Mode
		if { [info exists Mode($BandStack($ch,mode))] } \
			{
			set m $Mode($BandStack($ch,mode))
			} \
		else \
			{
			set m $Mode(NFM)
			}
		set b [SetBitField $b 6 7 $m]


		set image [string replace $image $first $first $b]
		incr first 1
		}


	return $image
}


###################################################################
# Encode the Call Channels frequency and mode
# information into a memory image.
###################################################################

proc EncodeCallChannels { image } \
{
	global CallChannel
	global Ctcss
	global ImageAddr
	global Mode
	global RToneFlag
	global Skip
	global Step

	global ImageAddr

	# Encode frequency.
	scan $ImageAddr(CallChannelFreq) "%x" first
	set last [expr {$first + 2}]

	for {set ch 0} {$ch < 2} {incr ch} \
		{
		set b [ Freq2BCD3 $CallChannel($ch,freq) ]

		set image [string replace $image $first $last $b]

		incr first 8
		incr last 8
		}

	# Encode offset frequencies.
	# and duplex/simple flags.
	scan $ImageAddr(CallChannelOffset) "%x" first
	set last [expr {$first + 2}]

	for {set ch 0} {$ch < 2} {incr ch} \
		{
		set b [ Offset2BCD $CallChannel($ch,offset) \
			$CallChannel($ch,duplex) ]

		set image [string replace $image $first $last $b]

		incr first 8
		incr last 8
		}



	# Encode channel rcv CTCSS code.
	scan $ImageAddr(CallChannelToneCode) "%x" first
	set next $first
	incr next

	for {set ch 0} {$ch < 2} {incr ch} \
		{
		set b0 [ string index $image $first ]
		set b1 [ string index $image $next ]

		# CTCSS rcv code
		set ccode $CallChannel($ch,tonecoder)

		if {$ccode != ""} \
			{
			# If pl code is a whole number
			if { ([string first "." $ccode] < 0) } \
				{
				append ccode .0
				}
			}


		if { [info exists Ctcss($ccode)] == 0 } \
			{
			set nr 0
			} \
		else \
			{
			set nr $Ctcss($ccode)
			}

		set b0 [SetBitField $b0 0 5 $nr]


		# CTCSS xmit code
		set ccode $CallChannel($ch,tonecodet)


		if {$ccode != ""} \
			{
			# If pl code is a whole number
			if { ([string first "." $ccode] < 0) } \
				{
				append ccode .0
				}
			}


		if { [info exists Ctcss($ccode)] == 0 } \
			{
			set nt 0
			} \
		else \
			{
			set nt $Ctcss($ccode)
			}

		set n0 [expr {$nt / 16}]
		set n0 [expr {int($n0)}]
		set b0 [SetBitField $b0 6 7 $n0]


		set n1 [expr {fmod($nt, 16)}]
		set n1 [expr {int($n1)}]
		set b1 [SetBitField $b1 0 3 $n1]

		set a0 [DumpBinary $b0]
		set a1 [DumpBinary $b1]

		set image [string replace $image $first $first $b0]
		set image [string replace $image $next $next $b1]

		incr first 8
		incr next 8
		}

	scan $ImageAddr(CallChannelSteps) "%x" first

	for {set ch 0} {$ch < 2} {incr ch} \
		{
		set b [ string index $image $first ]

		# Tuning Step

		if { [info exists Step($CallChannel($ch,step))] == 0 } \
			{
			set m 0
			} \
		else \
			{
			set m $Step($CallChannel($ch,step))
			}
		set b [SetBitField $b 4 7 $m]

		set image [string replace $image $first $first $b]
		incr first 8
		}


	# Encode Skip, Tone Squelch Flag, Duplex, Mode
	scan $ImageAddr(CallChannelSkip) "%x" first

	for {set ch 0} {$ch < 2} {incr ch} \
		{
		set b [ string index $image $first ]

		# Skip type
		if { [info exists Skip($CallChannel($ch,skip))] == 0 } \
			{
			set m 0
			} \
		else \
			{
			set m $Skip($CallChannel($ch,skip))
			}

		set b [SetBitField $b 4 5 $m]

		# Tone Squelch flag
		if { [info exists RToneFlag($CallChannel($ch,toneflag))] } \
			{
			set m $RToneFlag($CallChannel($ch,toneflag))
			} \
		else \
			{
			set m $RToneFlag(off)
			}

		set b [SetBitField $b 0 1 $m]

		if { [info exists Mode($CallChannel($ch,mode))] } \
			{
			set m $Mode($CallChannel($ch,mode))
			} \
		else \
			{
			set m $Mode(NFM))
			}
		set b [SetBitField $b 6 7 $m]

		# Duplex
		if {$CallChannel($ch,duplex) == "+"} \
			{
			set m 3
			} \
		elseif {$CallChannel($ch,duplex) == "-"} \
			{
			set m 2
			} \
		else \
			{
			set m 0
			}

		set b [SetBitField $b 2 3 $m]

		set image [string replace $image $first $first $b]
		incr first 1
		}


	return $image
}


###################################################################
# Encode the limit search bank channel numbers
# into a memory image.
###################################################################

proc EncodeSearchBanks { image } \
{
	global ImageAddr
	global LimitScan

	global ImageAddr

	# Encode lower channel.
	scan $ImageAddr(SearchChanFirst) "%x" first

	for {set bn 0} {$bn < 20} {incr bn} \
		{
		set s [format "%02x" $LimitScan($bn,lower)]
		set b [binary format "H2" $s]
		set image [string replace $image $first $first $b]
		incr first 2
		}


	# Encode upper channel.
	scan $ImageAddr(SearchChanSecond) "%x" first

	for {set bn 0} {$bn < 20} {incr bn} \
		{
		set s [format "%02x" $LimitScan($bn,upper)]
		set b [binary format "H2" $s]
		set image [string replace $image $first $first $b]
		incr first 2
		}

	return $image
}


###################################################################
# Pop up a window which says "Please wait..."
###################################################################
proc MakeWait { } \
{
	global DisplayFontSize


	toplevel .wait

	set w .wait
	wm title $w "tk7 running"

	label $w.lab -font $DisplayFontSize -text "Please wait ..."

	pack $w.lab

	update idletasks
	waiter 500
	return
}

###################################################################
# Kill the window which says "Please wait..."
###################################################################
proc KillWait { } \
{
	catch {destroy .wait}
	update idletasks
}

###################################################################
# ValidateData tests the data.
# It pops up a window with error and/or warning messages.
# If there are warnings but no errors, the user can elect
# to continue or cancel the current operation.
#
# Returns:
#	0	- continue
#	1	- cancel the current operation
###################################################################
proc ValidateData { } \
{
	global Band
	global BandStack
	global CallChannel
	global Emsg
	global GlobalParam
	global MemFreq
	global MemMode
	global MemSkip
	global LimitScan
	global Priority
	global Skip

	if { [info exists MemFreq(0)] == 0 } \
		{
		# No data to validate.
		return 1
		}


	set Emsg ""
	set nerror 0
	set nwarning 0

	# Memory channels.
	set ch 0
	for {set ch 0} {$ch < 200} {incr ch} \
		{
		set m "Bank Ch. $ch"
		set f $MemFreq($ch)

		if {$MemSkip($ch) == "hide"} \
			{
			continue
			}
		set code [ValidateFreq $f $m]
		if {$code == "error"} { incr nerror } \
		elseif {$code == "warning"} { incr nwarning }

		set f $MemMode($ch)
		set code [ValidateMode $f $m]
		if {$code == "error"} { incr nerror }

		if { [expr {$nerror + $nwarning}] > 5} {break}
		}




	# Limit scan
	for {set i 0} {$i < 20} {incr i} \
		{
		set m "Limit Search bank $i P lower"
		set ch $LimitScan($i,lower)

		set code [ValidateChNum $ch $m]
		if {$code == "error"} { incr nerror } \
		elseif {$code == "warning"} { incr nwarning }

		set m "Limit Search bank $i P upper"
		set ch $LimitScan($i,upper)

		set code [ValidateChNum $ch $m]
		if {$code == "error"} { incr nerror } \
		elseif {$code == "warning"} { incr nwarning }

		if { [expr {$nerror + $nwarning}] > 5} {break}
		}



	# BandStack frequencies
	for {set i 0} {$i < 8} {incr i} \
		{
		set m "Band Stacking "
		append m "$Band(USA,$i,low) - $Band(USA,$i,high)"
		set f $BandStack($i,freq)

		set code [ValidateFreq $f $m]
		if {$code == "error"} { incr nerror } \
		elseif {$code == "warning"} { incr nwarning }

		if { [expr {$nerror + $nwarning}] > 5} {break}
		}



	# CallChannel frequencies
	for {set i 0} {$i < 2} {incr i} \
		{
		set m "Call Channel $i"
		set f $CallChannel($i,freq)

		set code [ValidateFreq $f $m]
		if {$code == "error"} { incr nerror } \
		elseif {$code == "warning"} { incr nwarning }

		if { [expr {$nerror + $nwarning}] > 5} {break}
		}


	# Shortwave reception
		
	set m "Shortwave reception below 30 MHz warning:"
	set code [ValidateSW $m]
	if {$code == "error"} { incr nerror } \
	elseif {$code == "warning"} { incr nwarning }



	if {$nerror} \
		{
		tk_dialog .baddata1 "tk7 Invalid data" \
			$Emsg error 0 OK
		return 1
		}
	if {$nwarning} \
		{
		set response [tk_dialog .baddata2 \
			"tk7 Data warning" \
			$Emsg error 0 Cancel Continue]

		if {$response == 0} then {return 1} \
		else {return 0}
		}
	
	return 0
}

###################################################################
# Check a channel number for validity.
# Append the error message to a global string.
#
# Returns:
#	0
#	error
###################################################################
proc ValidateChNum { ch m } \
{
	global Emsg

	set code 0
	set msg ""

	if { ($ch == "") || ($ch < 0) || ($ch > 199) } \
		{
		append Emsg $m "\n"
		append Emsg "Channel number $ch is not between 0 and 199.\n"
		set code error
		}
	return $code
}


###################################################################
# Warn user if Shortwave reception below 30 MHz
# is enabled and channel 129 is not hidden.
# Append the warning message to a global string.
#
# Returns:
#	0
#	error
###################################################################
proc ValidateSW { m } \
{
	global Emsg
	global GlobalParam
	global MemSkip

	set code 0
	set msg ""

	if { $GlobalParam(SWenable) && ($MemSkip(129) != "hide") } \
		{
		append Emsg $m "\n"
		append Emsg "Enabling reception below 30 MHz. "
		append Emsg "will change the settings for "
		append Emsg "memory channel 129."
		set code warning
		}
	return $code
}

###################################################################
# Check a frequency for validity.
# Append the error or warning message to a global string.
#
# Returns:
#	0
#	warning
#	error
###################################################################
proc ValidateFreq {f m} \
{
	global Emsg
	global GlobalParam

	set code 0
	set msg ""

	if {( ($f != "") && ($f != 0.0) ) \
		&& (($f < $GlobalParam(LowestFreq)) \
		|| ($f > $GlobalParam(HighestFreq)))} \
		{
		append msg "\nError: $m frequency ($f) is out"
		append msg " of range.\n"
		set code error
		}


	append Emsg $msg
	return $code
}

###################################################################
# Return 1 if a string consists of 2 hex digits.
###################################################################
proc IsHex { s } \
{
	# Check for non-digit and non decimal point chars.
	set rc [regexp -nocase {^[0-9a-f][0-9a-f]$} $s]
	if {$rc} \
		{
		return 1
		} \
	else \
		{
		return 0
		}
}

###################################################################
# Check a mode for validity.
# Append the error message to a global string.
#
# Returns:
#	0
#	warning
#	error
###################################################################
proc ValidateMode {mode m} \
{
	global Emsg
	global Mode

	set code 0

	if { [info exists Mode($mode)] == 0} \
		{
		append Emsg "\nError: $m mode ($mode) is invalid.\n"
		set code error
		} 

	return $code
}

###################################################################
# Set title of the main window so it contains the
# current template file name.
###################################################################
proc SetWinTitle { } \
{
	global GlobalParam

	if { ( [info exists GlobalParam(TemplateFilename)] == 0 ) \
		|| ($GlobalParam(TemplateFilename) == "") } \
		{
		set filename untitled.tr2
		} \
	else \
		{
		set filename $GlobalParam(TemplateFilename)
		}

	set s [format "%s - tk7" $filename]
	wm title . $s

	return
}


# Prevent user from shrinking or expanding window.

proc FixSize { w } \
{
	wm minsize $w [winfo width $w] [winfo height $w]
	wm maxsize $w [winfo width $w] [winfo height $w]

	return
}


######################################################################
#					Bob Parnass
#					DATE:
#
# USAGE:	SortaBank first last
#
# INPUTS:
#		first	-starting channel to sort
#		last	-ending channel to sort
#
# RETURNS:
#		0	-ok
#		-1	-error
#
#
# PURPOSE:	Sort a range of memory channels based on frequency.
#
# DESCRIPTION:
#
######################################################################
proc SortaBank { first last } \
{
	global MemDuplex
	global MemFreq
	global MemMode
	global MemOffset
	global MemSkip
	global MemStep
	global MemToneCodeR
	global MemToneCodeT
	global MemToneFlag
	global Mimage


	if {[info exists Mimage] == 0} \
		{
		set msg "You must open a template file\n"
		append msg " or read an image from the radio\n"
		append msg " before sorting channels.\n"

		tk_dialog .belch "tk7" \
			$msg info 0 OK
		return -1
		}

	set inlist [Bank2List MemFreq $first $last]
	set vorder [SortFreqList $inlist]

	set slist [ReorderList $inlist $vorder]
	for {set i $first; set j 0} {$i <= $last} {incr i; incr j} \
		{
		set MemFreq($i) [lindex $slist $j]
		}

	set inlist [Bank2List MemMode $first $last]
	set slist [ReorderList $inlist $vorder]
	for {set i $first; set j 0} {$i <= $last} {incr i; incr j} \
		{
		set MemMode($i) [lindex $slist $j]
		if {$MemMode($i) == ""} \
			{
			set MemMode($i) NFM
			}
		}

	set inlist [Bank2List MemDuplex $first $last]
	set slist [ReorderList $inlist $vorder]
	for {set i $first; set j 0} {$i <= $last} {incr i; incr j} \
		{
		set MemDuplex($i) [lindex $slist $j]
		}

	set inlist [Bank2List MemOffset $first $last]
	set slist [ReorderList $inlist $vorder]
	for {set i $first; set j 0} {$i <= $last} {incr i; incr j} \
		{
		set MemOffset($i) [lindex $slist $j]
		}

	set inlist [Bank2List MemSkip $first $last]
	set slist [ReorderList $inlist $vorder]
	for {set i $first; set j 0} {$i <= $last} {incr i; incr j} \
		{
		set MemSkip($i) [lindex $slist $j]
		if {$MemSkip($i) == ""} \
			{
			set MemSkip($i) "hide"
			}
		}

	set inlist [Bank2List MemToneCodeT $first $last]
	set slist [ReorderList $inlist $vorder]
	for {set i $first; set j 0} {$i <= $last} {incr i; incr j} \
		{
		set MemToneCodeT($i) [lindex $slist $j]
		}



	set inlist [Bank2List MemToneCodeR $first $last]
	set slist [ReorderList $inlist $vorder]
	for {set i $first; set j 0} {$i <= $last} {incr i; incr j} \
		{
		set MemToneCodeR($i) [lindex $slist $j]
		}


	set inlist [Bank2List MemToneFlag $first $last]
	set slist [ReorderList $inlist $vorder]
	for {set i $first; set j 0} {$i <= $last} {incr i; incr j} \
		{
		set MemToneFlag($i) [lindex $slist $j]
		}


	set inlist [Bank2List MemStep $first $last]
	set slist [ReorderList $inlist $vorder]
	for {set i $first; set j 0} {$i <= $last} {incr i; incr j} \
		{
		set MemStep($i) [lindex $slist $j]
		}

	return 0
}

proc ClearAllChannels { } \
{
	global GlobalParam
	global Mimage

	global MemDuplex
	global MemFreq
	global MemMode
	global MemOffset
	global MemSkip
	global MemStep
	global MemToneCodeR
	global MemToneCodeT
	global MemToneFlag

	if { ([info exists Mimage] == 0) \
		|| ([string length $Mimage] <= 0)} \
		{
		# No image to write.
		set msg "You must first read template data from"
		append msg " the radio or open a file before"
		append msg " clearing memories."

		tk_dialog .error "Clear all channels" $msg error 0 OK
		return
		}


	set msg "Warning: This operation will clear all 200 "
	append msg "memory channels."

	set result [tk_dialog .clearall "Warning" \
		$msg warning 0 Cancel "Clear Memories" ]

	if {$result == 0} {return}

	for {set ch 0} {$ch < 200} {incr ch} \
		{
		ZapChannel $ch
		}

	return

}

proc ZapChannel { ch } \
{
	global MemDuplex
	global MemFreq
	global MemMode
	global MemOffset
	global MemSkip
	global MemStep
	global MemToneCodeR
	global MemToneCodeT
	global MemToneFlag


	set MemFreq($ch) 0.0
	set MemMode($ch) ?
	set MemOffset($ch) 0
	set MemToneCodeR($ch) 88.5
	set MemToneCodeT($ch) 88.5
	set MemDuplex($ch) "+"
	set MemStep($ch) 5
	set MemToneFlag($ch) tsql
	set MemSkip($ch) "hide"

	return
}
###################################################################
# Create memory banks. 
###################################################################
proc MakeMemoryBankFrame { f }\
{
	global GlobalParam
	global MemNB

	frame $f -relief groove -borderwidth 3
	# frame $f.b -relief flat -borderwidth 3

	label $f.lab -text "Memory Bank Settings" \
		-borderwidth 3


	pack $f.lab \
		-side top -fill both -expand false -padx 3 -pady 3

	MakeMemoryBankFrameCommon $f.common
	pack $f.common  -side left -padx 3 -pady 3 \
		-fill both

	set MemNB $f.banknb
	MakeMemoryBankNB $MemNB
	return $f
}

###################################################################
# Make a frame for settings common to all memory banks.
###################################################################
proc MakeMemoryBankFrameCommon { f }\
{
	global GlobalParam

	frame $f -relief groove -borderwidth 3

	return $f
}

###################################################################
# Make a single tabbed notebook to hold the settings
# for the individual memory banks.
###################################################################
proc MakeMemoryBankNB { w } \
{
	global NBanks

	tabnotebook_create $w

	for {set i 0} {$i < $NBanks} {incr i} \
		{
		set p [tabnotebook_page $w "Bank $i"]
		set fr $p.f; MakeMemoryBankPage $fr $i
		pack $fr
		}

	pack $w -expand true -fill both
}

###################################################################
# Make a frame to hold the settings
# for one memory bank.
###################################################################
proc MakeMemoryBankPage { f bn }\
{
	global GlobalParam

	frame $f -relief flat -borderwidth 3

	MakeBankSettingsFrame $f.c1 $bn
	MakeMemoryChannelFrame $f.c2 $bn
	MakeFillerFrame $f.c3 $bn

	pack $f.c1 $f.c2 $f.c3 -side left \
		-fill both -padx 3 -pady 3 -expand true

	return $f
}


###################################################################
# Create one a set of widgets which pertain to
# for one memory bank. 
#
# INPUTS:
#		f	-frame to create
#		bn	-bank number
#
###################################################################
proc MakeBankSettingsFrame { f bn }\
{
	global DisplayFontSize 

	frame $f -relief flat -borderwidth 3


	label $f.bankn$bn -text "$bn" \
		-relief flat \
		-borderwidth 6 \
		-font $DisplayFontSize 


	grid $f.bankn$bn -row 5 -column 0 -sticky ew 

	return $f
}


###################################################################
#
# INPUTS:
#		f	-frame to create
#		j	-index number
#
###################################################################
proc MakeFillerFrame { f j }\
{

	frame $f -relief flat -borderwidth 3


	for {set i 0} {$i < 12} {incr i} \
		{
		label $f.filler$j$i -text "-"  -relief flat \
			-borderwidth 6

		grid $f.filler$j$i -row $i -column 0 -sticky ew 
		}

	return $f
}


###################################################################
# Create widgets for memory channels for a bank. 
###################################################################
proc MakeMemoryChannelFrame { f bn }\
{
	global NBanks
	global NChanPerBank
	global VNChanPerBank

	frame $f -relief flat -borderwidth 3
	label $f.lab -text "Memory Channels" -borderwidth 3

	pack $f.lab -side top


	ScrollformCreate $f.b
	pack $f.b -expand yes -fill both

	set w [ScrollFormInterior $f.b]


	label $w.freq -text "Freq" -borderwidth 3
	label $w.mode -text "Mode" -borderwidth 3
	label $w.step -text "Step" -borderwidth 3
	label $w.offset -text "Offset" -borderwidth 3
	label $w.duplex -text "Duplex" -borderwidth 3
	label $w.toneflag -text "TSQL" -borderwidth 3
	label $w.tonecodet -text "TxCTCSS" -borderwidth 3
	label $w.tonecoder -text "RxCTCSS" -borderwidth 3
	label $w.skip -text "Skip" -borderwidth 3

	set ch [expr {$bn * $VNChanPerBank}]
	for {set i 0} {$i < $NChanPerBank} {incr i} \
		{
		MakeChannel $w $bn $i $ch
		incr ch
		update idletasks
		}

	grid $w.freq -row 0 -column 20
	grid $w.mode -row 0 -column 30
	grid $w.step -row 0 -column 40
	grid $w.offset -row 0 -column 50
	grid $w.duplex -row 0 -column 60
	grid $w.toneflag -row 0 -column 70
	grid $w.tonecodet -row 0 -column 80
	grid $w.tonecoder -row 0 -column 90
	grid $w.skip -row 0 -column 100

	return $f
}


###################################################################
# Create one a set of widgets for one channel. 
###################################################################
proc MakeChannel { f bn n ch }\
{
	global GlobalParam

	global MemDuplex
	global MemFreq
	global MemMode
	global MemOffset
	global MemSkip
	global MemStep
	global MemToneCodeR
	global MemToneCodeT
	global MemToneFlag

	set row [expr {$n + 2}]


	label $f.lab$bn$ch -text "$ch" -borderwidth 3

	entry $f.freq$bn$ch -width 12 \
		-textvariable MemFreq($ch) \
		-background white 

	tk_optionMenu $f.modemenu$bn$ch MemMode($ch) \
		NFM WFM AM

	tk_optionMenu $f.stepmenu$bn$ch MemStep($ch) \
		5 6.25 9 10 12.5 15 20 25 30 50 100


	entry $f.offset$bn$ch -width 12 \
		-textvariable MemOffset($ch) \
		-background white 

	tk_optionMenu $f.duplexmenu$bn$ch MemDuplex($ch) \
		"-" "+"

	tk_optionMenu $f.toneflagmenu$bn$ch MemToneFlag($ch) \
		off tone tsql pbeep

	entry $f.tonecodet$bn$ch -width 7 \
		-textvariable MemToneCodeT($ch) \
		-background white 

	entry $f.tonecoder$bn$ch -width 7 \
		-textvariable MemToneCodeR($ch) \
		-background white 

	tk_optionMenu $f.skipmenu$bn$ch MemSkip($ch) \
		pskip skip hide " "



	button $f.lower$bn$ch -text "^" \
		-command "MoveChannel $ch [expr {$ch - 1}]"

	button $f.higher$bn$ch -text "v" \
		-command "MoveChannel $ch [expr { $ch + 1}]"

	button $f.insert$bn$ch -text "Insert" \
		-command "InsertChannel $ch"

	button $f.delete$bn$ch -text "Delete" \
		-command "DeleteChannel $ch"

	grid $f.lab$bn$ch -row $row -column 10
	grid $f.freq$bn$ch -row $row -column 20
	grid $f.modemenu$bn$ch -row $row -column 30 -sticky ew
	grid $f.stepmenu$bn$ch -row $row -column 40 -sticky ew
	grid $f.offset$bn$ch -row $row -column 50 -sticky ew
	grid $f.duplexmenu$bn$ch -row $row -column 60 -sticky ew
	grid $f.toneflagmenu$bn$ch -row $row -column 70 -sticky ew
	grid $f.tonecodet$bn$ch -row $row -column 80 -sticky ew
	grid $f.tonecoder$bn$ch -row $row -column 90 -sticky ew
	grid $f.skipmenu$bn$ch -row $row -column 100 -sticky ew
	grid $f.lower$bn$ch -row $row -column 110
	grid $f.higher$bn$ch -row $row -column 120
	grid $f.insert$bn$ch -row $row -column 130
	grid $f.delete$bn$ch -row $row -column 140

	return
}

###################################################################
# Insert a memory channel and move all the higher channels
# in the same bank higher by one channel.  Clear the current
# channel in the bank.
###################################################################
proc InsertChannel { ch } \
{
	global MemDuplex
	global MemFreq
	global MemMode
	global MemOffset
	global MemSkip
	global MemStep
	global MemToneCodeR
	global MemToneCodeT
	global MemToneFlag

	global NChanPerBank
	global VNChanPerBank

	set bn [expr {int($ch/$VNChanPerBank)}]
	set last [expr {($bn * $VNChanPerBank) + $NChanPerBank - 1}]

	if {($MemFreq($last) != "") \
		|| ($MemFreq($last) > 0)} \
		{
		# No room.

		set msg "Channel $last is not empty.\n\n"
		append msg "Please delete channel $last before "
		append msg "inserting a new channel $ch and "
		append msg "moving the existing channels higher."

		tk_dialog .belch "Insert new channel" \
			$msg error 0 OK

		return
		}

	set n [expr {$NChanPerBank - fmod($ch, $VNChanPerBank) - 1}]

	set to $last
	set from $last
	incr from -1

	for {set i 0} {$i < $n} {incr i} \
		{
		# puts stderr "InsertChannel: n: $n, moving channel $from to $to"
		set MemFreq($to) $MemFreq($from)
		set MemMode($to) $MemMode($from)
		set MemStep($to) $MemStep($from)
		set MemOffset($to) $MemOffset($from)
		set MemDuplex($to) $MemDuplex($from)
		set MemToneFlag($to) $MemToneFlag($from)
		set MemToneCodeT($to) $MemToneCodeT($from)
		set MemToneCodeR($to) $MemToneCodeR($from)
		set MemSkip($to) $MemSkip($from)

		incr from -1
		incr to	-1
		}

	ZapChannel $ch

}

###################################################################
# Delete a memory channel and move all the higher channels
# in the same bank to the previous channel.  Clear the last
# channel in the bank.
###################################################################

proc DeleteChannel { ch } \
{

	global MemDuplex
	global MemFreq
	global MemMode
	global MemOffset
	global MemSkip
	global MemStep
	global MemToneCodeR
	global MemToneCodeT
	global MemToneFlag

	global NChanPerBank
	global VNChanPerBank

	set n [expr {$NChanPerBank - fmod($ch, $VNChanPerBank) - 1}]

	set to $ch
	set from $ch
	incr from

	for {set i 0} {$i < $n} {incr i} \
		{
		# puts stderr "DeleteChannel: moving channel $from to $to"
		set MemFreq($to) $MemFreq($from)
		set MemMode($to) $MemMode($from)
		set MemStep($to) $MemStep($from)
		set MemOffset($to) $MemOffset($from)
		set MemDuplex($to) $MemDuplex($from)
		set MemToneFlag($to) $MemToneFlag($from)
		set MemToneCodeT($to) $MemToneCodeT($from)
		set MemToneCodeR($to) $MemToneCodeR($from)
		set MemSkip($to) $MemSkip($from)

		incr from	
		incr to	
		}

	set bn [expr {int($ch/$VNChanPerBank)}]
	set last [expr {($bn * $VNChanPerBank) + $NChanPerBank - 1}]

	ZapChannel $last

	return
}


###################################################################
# Swap channel with the next higher channel in the bank.
###################################################################
proc MoveChannel { ch1 ch2 } \
{
	global MemDuplex
	global MemFreq
	global MemMode
	global MemOffset
	global MemSkip
	global MemStep
	global MemToneCodeR
	global MemToneCodeT
	global MemToneFlag

	global NChanPerBank
	global VNChanPerBank

	set bn [expr {int($ch1/$VNChanPerBank)}]
	set last [expr {($bn * $VNChanPerBank) + $NChanPerBank - 1}]
	set ch1r [expr {int(fmod($ch1,$VNChanPerBank))}]
	set ch2r [expr {int(fmod($ch2,$VNChanPerBank))}]

	if {($ch1 > $last) \
		|| ($ch1 < 0) \
		|| ($ch2 > $last) \
		|| ($ch2 < 0) \
		|| ($ch1r >= $NChanPerBank) \
		|| ($ch2r >= $NChanPerBank) } \
		{

		set msg "Cannot move channel $ch1 to $ch2."

		tk_dialog .belch "Move channel" \
			$msg error 0 OK

		return
		}

	set tmp $MemFreq($ch2)
	set MemFreq($ch2) $MemFreq($ch1)
	set MemFreq($ch1) $tmp

	set tmp $MemMode($ch2)
	set MemMode($ch2) $MemMode($ch1)
	set MemMode($ch1) $tmp

	set tmp $MemStep($ch2)
	set MemStep($ch2) $MemStep($ch1)
	set MemStep($ch1) $tmp

	set tmp $MemOffset($ch2)
	set MemOffset($ch2) $MemOffset($ch1)
	set MemOffset($ch1) $tmp

	set tmp $MemDuplex($ch2)
	set MemDuplex($ch2) $MemDuplex($ch1)
	set MemDuplex($ch1) $tmp

	set tmp $MemToneFlag($ch2)
	set MemToneFlag($ch2) $MemToneFlag($ch1)
	set MemToneFlag($ch1) $tmp

	set tmp $MemToneCodeT($ch2)
	set MemToneCodeT($ch2) $MemToneCodeT($ch1)
	set MemToneCodeT($ch1) $tmp

	set tmp $MemToneCodeR($ch2)
	set MemToneCodeR($ch2) $MemToneCodeR($ch1)
	set MemToneCodeR($ch1) $tmp

	set tmp $MemSkip($ch2)
	set MemSkip($ch2) $MemSkip($ch1)
	set MemSkip($ch1) $tmp

	return
}

