#!/bin/sh
# the next line restarts using wish \
exec wish "$0" "$@"

###################################################################
# This file is part of tk7, a utility program for the
# ICOM IC-Q7A dual band walkie-talkie.
# 
#    Copyright (C) 2001, 2002, Bob Parnass
#					AJ9S
# 
# tk7 is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
# 
# tk7 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with tk7; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# 02111-1307  USA
###################################################################


######################################################################
# NOTES:
#	There are 2 levels of data structures within this
#	program:
#
#	(1) Mimage - a 1984 byte string stored in binary format
#		It must be converted to another form for reading and
#		writing to the radio.
#
#	(2) Arrays containing memory channel and configuration
#		information.
#
#
#	The memory image Mimage string becomes populated in these
#	scenarios:
#
#		- image data is read from the radio
#		- image data is read from a .tk7 file
#		- array data is encoded into the image
#			prior to writing the image to the radio
#
#	Some or all of the arrays become populated in these
#	scenarios:
#
#		- by parsing the memory image string
#		- by importing memory channel info from a .csv file
#		- by a user changing widget values in the GUI
#
# Most of the protocol and memory image layout information
# used in this software was obtained from the notes
# compiled by Vojtech Bubnik on Goran's Vlaski's original work
# and a document entitled
# "Cloning ICOM Receivers, Using the ICOM IC-R2 Handheld
# Scanner as an example," by BlakkeKatte@yahoo.co.uk.
# It is available for download at http://uk.geocities.com/blakkekatte
#
#
# A note by Vojtech Bubnik:
#     Super RX (RX from 0 to 1.8 GHz):
#     Set address 0764h to 15h (sets VFO to nonexistent
#     -- overflow to channel 129)
#     
#     Set memory channel 129 (addresses 0408 to 040F) to
#     03h 00h 00h 00h 00h 00h 20h 82h (30.0 MHz, TSQ 88.5,
#     tuning step 10.0 kHz)
#     
#     Set vfo data (addresses from 07B8 to 07BF, is unused) to
#          49h 63h 6Fh 6Dh 43h 6Ch 6Fh 6Eh
#          
#     Don't ask me, why. I don't know. It is Goran's hack, ask him.
#
######################################################################

######################################################################
# Write error messages to stderr if linux/unix, stdout otherwise
######################################################################

proc Tattle { msg } \
{
	global tcl_platform

	 set platform $tcl_platform(os) 
	 switch -glob $platform \
		{
		{[Ll]inux} \
			{
			puts stderr $msg
			}
		{unix} \
			{
			puts stderr $msg
			}
		default \
			{
			puts $msg
			}
		}

	return
}
############################################################
set Version "0.5"

set AboutMsg  "tk7
version $Version

Copyright 2001, 2002, Bob Parnass
Oswego, Illinois
USA
http://parnass.com

Released under the GNU General Public License.

tk7 is a utility program for the
ICOM IC-Q7 walkie-talkie.
This is beta software. If you find a defect,
please report it."

############################################################

# trace variable Sid r {puts stderr "Sid trace trap"}
set Pgm [lindex [split $argv0 "/"] end]
set Lfilename ""

if { [catch {set Libdir $env(tk7)} ] } \
	{
	Tattle "$Pgm: error: Environment variable tk7 must"
	Tattle "be set to the directory containing the library"
	Tattle "files for program $Pgm."
	exit 1
	}

set ScanFlag 0


# Sanity check for the Libdir environment variable.
if {$Libdir == ""}\
	{
	Tattle "$Pgm: error: Environment variable tk7 must"
	Tattle "be set to the directory containing the library"
	Tattle "files for program $Pgm."
	exit 1
	}

source [ file join $Libdir "misclib.tcl" ]
source [ file join $Libdir "mylib.tcl" ]
source [ file join $Libdir "api7.tcl" ]
source [ file join $Libdir "gui7.tcl" ]

SetUp


SetWinTitle

set lst [ InitStuff ]
set Rcfile [ lindex $lst 0 ]
set LabelFile [ lindex $lst 1 ]

FirstTimeCheck $Rcfile

# Set most global variables from configuration file.

PresetGlobals
ReadSetup
OverrideGlobals



set CancelXfer 0

set FileTypes \
	{
	{"IC-R2 data files"           {.csv .txt}     }
	}


# Create graphical widgets.

MakeGui

update idletasks
